from functools import partial
from typing import TYPE_CHECKING, Dict, Optional, Union

import torch

from eir.models.input.array.models_locally_connected import FlattenFunc, LCLModel
from eir.models.model_setup_modules.output_model_setup_modules import al_output_modules
from eir.models.output.array.array_output_modules import (
    ArrayOutputModuleConfig,
    CNNPassThroughUpscaleModel,
    CNNUpscaleModel,
    CNNUpscaleModelConfig,
    LCLOutputModelConfig,
    al_array_model_types,
    al_output_array_model_classes,
    al_output_array_model_config_classes,
    al_output_array_models,
    get_array_output_module,
)
from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.output_setup_modules.array_output_setup import ComputedArrayOutputInfo
from eir.setup.output_setup_modules.image_output_setup import ComputedImageOutputInfo
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import FeatureExtractorInfo

al_output_array_model_init_kwargs = dict[
    str,
    Union[
        Optional[DataDimensions],
        LCLOutputModelConfig | CNNUpscaleModelConfig,
        dict[str, "FeatureExtractorInfo"],
        FlattenFunc,
        int,
        str,
    ],
]

al_output_array_model_configs = LCLOutputModelConfig | CNNUpscaleModelConfig

logger = get_logger(name=__name__)


def get_array_or_image_output_module_from_model_config(
    output_object: ComputedArrayOutputInfo | ComputedImageOutputInfo,
    input_dimension: Optional[int],
    fusion_model_type: str,
    feature_extractor_infos: Optional[dict[str, "FeatureExtractorInfo"]],
    device: str,
) -> al_output_modules:
    output_model_config = output_object.output_config.model_config
    assert isinstance(output_model_config, ArrayOutputModuleConfig)

    output_module_type = output_model_config.model_type
    output_name = output_object.output_config.output_info.output_name

    output_module: al_output_modules

    if input_dimension is None:
        input_data_dimension = None
    else:
        input_data_dimension = DataDimensions(
            channels=1,
            height=1,
            width=input_dimension,
        )

    is_diffusion = output_object.diffusion_config is not None
    is_pass_through_fusion = fusion_model_type == "pass-through"
    use_passthrough = is_diffusion or is_pass_through_fusion

    diffusion_time_steps = None
    if is_diffusion:
        assert output_object.diffusion_config is not None
        diffusion_time_steps = output_object.diffusion_config.time_steps

    feature_extractor = get_array_output_feature_extractor(
        model_init_config=output_model_config.model_init_config,
        input_data_dimensions=input_data_dimension,
        feature_extractor_infos=feature_extractor_infos,
        model_type=output_module_type,
        output_data_dimensions=output_object.data_dimensions,
        output_name=output_name,
        use_passthrough=use_passthrough,
        diffusion_time_steps=diffusion_time_steps,
    )

    array_output_module = get_array_output_module(
        feature_extractor=feature_extractor,
        output_name=output_name,
        target_data_dimensions=output_object.data_dimensions,
    )

    torch_device = torch.device(device=device)
    output_module = array_output_module.to(device=torch_device)

    return output_module


def get_array_output_feature_extractor(
    model_init_config: al_output_array_model_configs,
    input_data_dimensions: Optional[DataDimensions],
    feature_extractor_infos: Optional[dict[str, "FeatureExtractorInfo"]],
    model_type: al_array_model_types,
    output_data_dimensions: Optional[DataDimensions],
    output_name: str,
    use_passthrough: bool,
    diffusion_time_steps: Optional[int],
) -> al_output_array_models:

    model_type = parse_model_type(
        model_type=model_type,
        use_passthrough=use_passthrough,
    )

    array_model_class = get_array_output_model_class(model_type=model_type)
    model_init_kwargs = get_array_output_model_init_kwargs(
        model_type=model_type,
        model_config=model_init_config,
        fused_input_data_dimensions=input_data_dimensions,
        feature_extractor_infos=feature_extractor_infos,
        output_data_dimensions=output_data_dimensions,
        output_name=output_name,
        diffusion_time_steps=diffusion_time_steps,
    )

    array_model = array_model_class(**model_init_kwargs)  # type: ignore
    return array_model


def parse_model_type(
    model_type: al_array_model_types, use_passthrough: bool
) -> al_array_model_types:
    if model_type == "cnn" and use_passthrough:
        return "cnn-passthrough"
    return model_type


def get_array_output_model_mapping() -> Dict[str, al_output_array_model_classes]:
    mapping = {
        "lcl": LCLModel,
        "cnn": CNNUpscaleModel,
        "cnn-passthrough": CNNPassThroughUpscaleModel,
    }

    return mapping


def get_array_output_model_class(
    model_type: al_array_model_types,
) -> al_output_array_model_classes:
    mapping = get_array_output_model_mapping()
    return mapping[model_type]


def get_array_output_config_dataclass_mapping() -> (
    Dict[str, al_output_array_model_config_classes]
):
    mapping = {
        "lcl": LCLOutputModelConfig,
        "cnn": CNNUpscaleModelConfig,
        "cnn-passthrough": CNNUpscaleModelConfig,
    }

    return mapping


def get_array_output_model_config_dataclass(
    model_type: str,
) -> al_output_array_model_config_classes:
    mapping = get_array_output_config_dataclass_mapping()
    return mapping[model_type]


def get_array_output_model_init_kwargs(
    model_type: al_array_model_types,
    model_config: al_output_array_model_configs,
    fused_input_data_dimensions: Optional[DataDimensions],
    feature_extractor_infos: Optional[dict[str, "FeatureExtractorInfo"]],
    output_data_dimensions: Optional[DataDimensions],
    output_name: str,
    diffusion_time_steps: Optional[int],
) -> al_output_array_model_init_kwargs:
    kwargs: al_output_array_model_init_kwargs = {}

    model_config_dataclass = get_array_output_model_config_dataclass(
        model_type=model_type
    )
    model_config_dataclass_instance = model_config_dataclass(**model_config.__dict__)

    kwargs["model_config"] = model_config_dataclass_instance

    match model_type:
        case "lcl":
            assert isinstance(model_config, LCLOutputModelConfig)
            assert fused_input_data_dimensions is not None
            kwargs["data_dimensions"] = fused_input_data_dimensions
            kwargs["flatten_fn"] = partial(torch.flatten, start_dim=1)

            if model_config.cutoff == "auto":
                assert output_data_dimensions is not None
                num_elements = output_data_dimensions.num_elements()
                logger.debug(
                    "Setting dynamic cutoff to %s for LCL array output module.",
                    num_elements,
                )
                kwargs["dynamic_cutoff"] = num_elements

        case "cnn":
            assert isinstance(model_config, CNNUpscaleModelConfig)
            assert output_data_dimensions is not None
            kwargs["target_dimensions"] = output_data_dimensions
            kwargs["data_dimensions"] = fused_input_data_dimensions
        case "cnn-passthrough":
            assert isinstance(model_config, CNNUpscaleModelConfig)
            assert feature_extractor_infos is not None
            kwargs["target_dimensions"] = output_data_dimensions
            kwargs["feature_extractor_infos"] = feature_extractor_infos
            kwargs["output_name"] = output_name

            if diffusion_time_steps:
                kwargs["diffusion_time_steps"] = diffusion_time_steps

    return kwargs



from dataclasses import asdict, dataclass, fields
from pathlib import Path
from typing import Dict, Generator, Optional, Sequence, Tuple

import numpy as np
import torch
from PIL import Image
from timm.models._registry import _model_pretrained_cfgs
from torchvision import transforms
from torchvision.transforms import Compose
from torchvision.transforms.functional import to_tensor

from eir.data_load.data_source_modules.deeplake_ops import (
    get_deeplake_input_source_iterable,
    is_deeplake_dataset,
    load_deeplake_dataset,
)
from eir.models.input.image.image_models import ImageModelConfig
from eir.models.output.array.array_output_modules import ArrayOutputModuleConfig
from eir.setup import schemas
from eir.setup.input_setup_modules.common import DataDimensions
from eir.setup.setup_utils import ChannelBasedRunningStatistics, collect_stats
from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


@dataclass
class PretrainedImageModelInfo:
    url: str
    num_classes: int
    input_size: Sequence[int]
    pool_size: Sequence[int]
    mean: Sequence[float]
    std: Sequence[float]
    first_conv: str
    classifier: str


def get_timm_configs() -> Dict[str, PretrainedImageModelInfo]:
    default_configs = {}
    field_names = {i.name for i in fields(PretrainedImageModelInfo)}
    for name, pretrained_config in _model_pretrained_cfgs.items():
        config_dict = asdict(pretrained_config)
        common = {k: v for k, v in config_dict.items() if k in field_names}

        default_configs[name] = PretrainedImageModelInfo(**common)

    return default_configs


@dataclass
class ComputedImageInputInfo:
    input_config: schemas.InputConfig
    base_transforms: Compose
    all_transforms: Compose
    normalization_stats: "ImageNormalizationStats"
    num_channels: int
    data_dimensions: "DataDimensions"


def set_up_image_input_for_training(
    input_config: schemas.InputConfig, *args, **kwargs
) -> ComputedImageInputInfo:
    input_type_info = input_config.input_type_info
    assert isinstance(input_type_info, schemas.ImageInputDataConfig)

    image_mode = input_type_info.mode
    num_channels = input_type_info.num_channels
    if not num_channels and not image_mode:
        num_channels = infer_num_image_channels(
            data_source=input_config.input_info.input_source,
            deeplake_inner_key=input_config.input_info.input_inner_key,
        )

    num_channels = get_num_channels_wrapper(
        image_mode=image_mode,
        num_channels=num_channels,
        source=input_config.input_info.input_source,
        deeplake_inner_key=input_config.input_info.input_inner_key,
    )

    data_dimension = DataDimensions(
        channels=num_channels,
        height=input_type_info.size[0],
        width=input_type_info.size[-1],
    )

    iti = input_type_info
    model_config = input_config.model_config
    assert isinstance(model_config, ImageModelConfig)
    normalization_stats = get_image_normalization_values(
        source=input_config.input_info.input_source,
        inner_key=input_config.input_info.input_inner_key,
        model_config=model_config,
        mean_normalization_values=iti.mean_normalization_values,
        stds_normalization_values=iti.stds_normalization_values,
        adaptive_normalization_max_samples=iti.adaptive_normalization_max_samples,
        data_dimensions=data_dimension,
        image_mode=iti.mode,
    )

    base_transforms, all_transforms = get_image_transforms(
        target_size=input_type_info.size,
        normalization_stats=normalization_stats,
        auto_augment=input_type_info.auto_augment,
        resize_approach=input_type_info.resize_approach,
    )

    image_input_info = ComputedImageInputInfo(
        input_config=input_config,
        base_transforms=base_transforms,
        all_transforms=all_transforms,
        normalization_stats=normalization_stats,
        num_channels=num_channels,
        data_dimensions=data_dimension,
    )

    return image_input_info


def get_num_channels_wrapper(
    image_mode: Optional[str],
    num_channels: Optional[int],
    source: str,
    deeplake_inner_key: Optional[str],
) -> int:

    if image_mode and num_channels:
        raise ValueError(
            "Got both image mode and number of channels. Please only specify one."
        )

    if image_mode is not None:
        num_channels = get_num_channels_from_image_mode(image_mode=image_mode)
        return num_channels

    elif num_channels:
        return num_channels

    else:
        num_channels = infer_num_image_channels(
            data_source=source,
            deeplake_inner_key=deeplake_inner_key,
        )
        return num_channels


def get_num_channels_from_image_mode(image_mode: Optional[str]) -> int:

    match image_mode:
        case "RGB":
            channels = 3
        case "L":
            channels = 1
        case "RGBA":
            channels = 4
        case _:
            raise ValueError(f"Unknown image mode: {image_mode}")

    logger.debug(
        "Inferred number of channels from image mode %s as: %d", image_mode, channels
    )

    return channels


def infer_num_image_channels(
    data_source: str, deeplake_inner_key: Optional[str]
) -> int:
    if is_deeplake_dataset(data_source=data_source):
        assert deeplake_inner_key is not None
        deeplake_ds = load_deeplake_dataset(data_source=data_source)
        deeplake_iter = get_deeplake_input_source_iterable(
            deeplake_dataset=deeplake_ds, inner_key=deeplake_inner_key
        )
        test_image_array = next(deeplake_iter).numpy()
        data_pointer = (
            f"[deeplake dataset {data_source}, input {deeplake_inner_key}, "
            f"image ID: {deeplake_ds['ID'][0].text()}]"
        )
    else:
        test_file = next(Path(data_source).iterdir())
        test_image = default_image_loader(path=str(test_file))
        test_image_array = np.array(test_image)
        data_pointer = test_file.name

    if test_image_array.ndim == 2:
        num_channels = 1
    else:
        num_channels = test_image_array.shape[-1]

    logger.info(
        "Inferring number of channels from source %s (using %s) as: %d",
        data_source,
        data_pointer,
        num_channels,
    )

    return num_channels


def default_image_loader(path: str) -> Image.Image:
    with open(path, "rb") as f:
        img = Image.open(f)
        img.load()
        return img


@dataclass
class ImageNormalizationStats:
    means: torch.Tensor
    stds: torch.Tensor


def get_image_normalization_values(
    source: str,
    inner_key: Optional[str],
    model_config: ImageModelConfig | ArrayOutputModuleConfig,
    mean_normalization_values: Optional[Sequence[float] | torch.Tensor],
    stds_normalization_values: Optional[Sequence[float] | torch.Tensor],
    adaptive_normalization_max_samples: Optional[int],
    data_dimensions: DataDimensions,
    image_mode: Optional[str],
) -> ImageNormalizationStats:

    pretrained_model_configs = get_timm_configs()

    means: Optional[torch.Tensor | Sequence[float]]
    stds: Optional[torch.Tensor | Sequence[float]]

    means = mean_normalization_values
    stds = stds_normalization_values

    if hasattr(model_config, "pretrained_model") and model_config.pretrained_model:
        cur_config = pretrained_model_configs[model_config.model_type]

        if not means:
            logger.info(
                "Using inferred image channel means (%s) from base on training "
                "statistics from pretrained '%s' model.",
                cur_config.mean,
                model_config.model_type,
            )
            means = cur_config.mean
        else:
            logger.warning(
                "Got manual values for channel means (%s) when using "
                "pretrained model '%s'. Usually one would use the means "
                "from the training data when '%s' was trained.",
                means,
                model_config.model_type,
                model_config.model_type,
            )
        if not stds:
            logger.info(
                "Using inferred image channel standard deviations (%s) from base on "
                "training statistics from pretrained '%s' model.",
                cur_config.std,
                model_config.model_type,
            )
            stds = cur_config.std
        else:
            logger.warning(
                "Got manual values for channel standard deviations (%s) "
                "when using pretrained model '%s'. Usually one would use "
                "the means from the training data when '%s' was trained.",
                stds,
                model_config.model_type,
                model_config.model_type,
            )
    else:
        if not means or not stds:
            input_source = source
            deeplake_inner_key = inner_key
            logger.info(
                "Not using an external pretrained model "
                "and no mean and standard deviation "
                "statistics passed in. Gathering running image means and standard "
                "deviations from %s.",
                input_source,
            )

            if is_deeplake_dataset(data_source=input_source):
                deeplake_ds = load_deeplake_dataset(data_source=input_source)
                assert deeplake_inner_key is not None
                image_iter = get_deeplake_input_source_iterable(
                    deeplake_dataset=deeplake_ds, inner_key=deeplake_inner_key
                )

                numpy_iter = (i.numpy() for i in image_iter)

                if image_mode:
                    image_iter = (Image.fromarray(i) for i in numpy_iter)
                    image_iter = (i.convert(image_mode) for i in image_iter)
                    numpy_iter = (np.array(i) for i in image_iter)

                tensor_iterator = (to_tensor(i).float() for i in numpy_iter)

            else:
                file_iterator = Path(input_source).rglob("*")
                image_iterator = (default_image_loader(str(f)) for f in file_iterator)
                if image_mode:
                    image_iterator = (i.convert(image_mode) for i in image_iterator)

                tensor_iterator = (to_tensor(i) for i in image_iterator)

            tensor_iterator = _get_maybe_truncated_tensor_iterator(
                tensor_iterator=tensor_iterator,
                max_samples=adaptive_normalization_max_samples,
            )

            gathered_stats = collect_stats(
                tensor_iterable=tensor_iterator,
                collector_class=ChannelBasedRunningStatistics,
                shape=data_dimensions.full_shape(),
            )
            means = gathered_stats.mean
            stds = gathered_stats.std
            logger.info(
                "Gathered the following means: %s and standard deviations: %s "
                "from %s.",
                means,
                stds,
                input_source,
            )

    assert means is not None
    assert stds is not None

    if torch.is_tensor(means):
        assert isinstance(means, torch.Tensor)
        means_tensor = means.clone().detach()
    else:
        means_tensor = torch.tensor(means)
    if torch.is_tensor(stds):
        assert isinstance(stds, torch.Tensor)
        stds_tensor = stds.clone().detach()
    else:
        stds_tensor = torch.tensor(stds)

    stats = ImageNormalizationStats(
        means=means_tensor,
        stds=stds_tensor,
    )

    return stats


def _get_maybe_truncated_tensor_iterator(
    tensor_iterator: Generator[torch.Tensor, None, None], max_samples: Optional[int]
) -> Generator[torch.Tensor, None, None]:

    if max_samples is not None:
        tensor_iterator = (t for _, t in zip(range(max_samples), tensor_iterator))

    return tensor_iterator


def get_image_transforms(
    target_size: Sequence[int],
    normalization_stats: ImageNormalizationStats,
    auto_augment: bool,
    resize_approach: str,
) -> Tuple[transforms.Compose, transforms.Compose]:
    if len(target_size) == 1:
        target_size = (target_size[0], target_size[0])
        logger.info(
            "Got target size as a single value. Assuming square image and "
            "setting target size to %s.",
            target_size,
        )

    random_transforms = transforms.TrivialAugmentWide()

    if resize_approach == "randomcrop":
        resize_transform = transforms.Resize([int(i * 1.5) for i in target_size])
        crop_transform = transforms.RandomCrop(target_size)
    elif resize_approach == "centercrop":
        resize_transform = transforms.Resize([int(i * 1.5) for i in target_size])
        crop_transform = transforms.CenterCrop(target_size)
    else:
        resize_transform = transforms.Resize(target_size)
        crop_transform = transforms.Lambda(lambda x: x)

    base = [
        resize_transform,
        crop_transform,
        transforms.ToTensor(),
        transforms.Normalize(
            mean=normalization_stats.means,
            std=normalization_stats.stds,
        ),
    ]

    base_transforms = transforms.Compose(transforms=base)
    if auto_augment:
        logger.info("Image will be auto augmented with TrivialAugment during training.")
        all_transforms = transforms.Compose(transforms=[random_transforms] + base)
    else:
        all_transforms = base_transforms

    return base_transforms, all_transforms



from copy import copy, deepcopy
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Sequence,
    Tuple,
    Union,
)

import matplotlib.pyplot as plt
import torch
from aislib.pytorch_modules import Swish
from ignite.engine import Engine
from ignite.handlers.lr_finder import FastaiLRFinder
from torch import nn
from torch.nn import Module
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader

from eir.setup.output_setup import ComputedTabularOutputInfo
from eir.train_utils.utils import call_hooks_stage_iterable
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from eir.setup.output_setup import al_output_objects_as_dict
    from eir.train import Experiment  # noqa: F401
    from eir.train_utils.step_logic import (
        al_ids,
        al_input_batch,
        al_training_labels_target,
    )

# Aliases
al_dataloader_gathered_predictions = Tuple[
    Dict[str, Dict[str, torch.Tensor]], "al_training_labels_target", List[str]
]
al_dataloader_gathered_raw = Tuple[
    Dict[str, torch.Tensor], "al_training_labels_target", Sequence[str]
]
al_lr_find_results = Dict[str, List[Union[float, List[float]]]]

logger = get_logger(name=__name__, tqdm_compatible=True)


def predict_on_batch(
    model: Module, inputs: Dict[str, torch.Tensor]
) -> Dict[str, Dict[str, torch.Tensor]]:
    assert not model.training
    with torch.no_grad():
        val_outputs = model(inputs=inputs)

    return val_outputs


class ColumnType(Enum):
    CON = "con"
    CAT = "cat"


def prepare_all_targets(
    output_objects: "al_output_objects_as_dict",
    device: str,
    labels: "al_training_labels_target",
) -> "al_training_labels_target":
    labels_prepared = parse_tabular_target_labels(
        output_objects=output_objects,
        device=device,
        labels=labels,
    )

    labels_prepared = recursive_to_device(obj=labels_prepared, device=device)

    return labels_prepared


def recursive_to_device(
    obj: Any,
    device: str,
) -> Any:
    if isinstance(obj, torch.Tensor):
        return obj.to(device=device)
    elif isinstance(obj, dict):
        return {
            key: recursive_to_device(obj=value, device=device)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [recursive_to_device(obj=value, device=device) for value in obj]
    else:
        return obj


def parse_tabular_target_labels(
    output_objects: "al_output_objects_as_dict",
    device: str,
    labels: "al_training_labels_target",
) -> "al_training_labels_target":
    labels_casted = copy(labels)

    def handle_tabular_object(
        output_object_: "ComputedTabularOutputInfo",
        output_name_: str,
    ) -> None:
        if output_name_ not in labels_casted:
            labels_casted[output_name_] = {}

        target_columns = output_object_.target_columns
        for column_type, list_of_cols_of_this_type in target_columns.items():
            for column_name in list_of_cols_of_this_type:
                cur_labels = labels[output_name_][column_name]

                if column_type == ColumnType.CON.value:
                    labels_casted[output_name_][column_name] = cur_labels.to(
                        dtype=torch.float, device=device
                    )
                elif column_type == ColumnType.CAT.value:
                    cur_labels = replace_nan_and_cast_to_long(
                        cur_labels=cur_labels.to(dtype=torch.float), device=device
                    )
                    labels_casted[output_name_][column_name] = cur_labels.to(
                        dtype=torch.long, device=device
                    )

    for output_name, output_object in output_objects.items():
        match output_object:
            case ComputedTabularOutputInfo():
                handle_tabular_object(
                    output_object_=output_object,
                    output_name_=output_name,
                )

    return labels_casted


def replace_nan_and_cast_to_long(
    cur_labels: torch.Tensor,
    device: str,
    replacement_value: int = -1,
) -> torch.Tensor:
    """
    Replace NaN values in a PyTorch tensor and cast it to a long integer
    (torch.long) type.

    When casting a tensor from a floating-point type (like float64) to an
    integer type (like long), NaN values need to be handled specially. In PyTorch,
    the behavior of casting NaN values to integers can be inconsistent
    across different environments, leading to unexpected results.

    For example:
    - On a local machine (e.g., Mac), NaN values may be converted to 0.
    - On some servers (e.g., GitHub Actions server), NaN values may be converted to
      -9223372036854775808, which is the minimum value for a 64-bit integer.

    This function explicitly replaces NaN values in the tensor with a defined integer
    before casting. This ensures consistent and defined behavior across
    different platforms.

    Example:
    >>> cur_labels = torch.tensor([nan, nan, 2., nan, 1.], dtype=torch.float32)
    >>> cur_labels.to(dtype=torch.long)
    tensor([0, 0, 2, 0, 1])
    or
    tensor([-9223372036854775808, -9223372036854775808, 2, -9223372036854775808, 1])
    >>> cur_labels = torch.tensor([nan, nan, 2., nan, 1.], dtype=torch.float32)
    >>> replace_nan_and_cast_to_long(cur_labels)
    tensor([-1, -1,  2, -1,  1])

    Note:
    The choice of `replacement_value` should be made carefully, in the case of
    categories, they are encoded from 0 to n-1, so using a negative value is
    potentially a reasonable choice.
    """

    replacement_tensor = torch.tensor(
        replacement_value,
        dtype=torch.float32,
        device=device,
    )

    cur_labels = cur_labels.to(device=device)

    cur_labels = cur_labels.where(~cur_labels.isnan(), replacement_tensor)
    return cur_labels.to(dtype=torch.long)


def get_prediction_outputs_generator(
    data_loader: DataLoader,
    batch_prep_hook: Iterable[Callable],
    batch_prep_hook_kwargs: Dict[str, Any],
    model: Module,
    with_labels: bool = True,
) -> Generator[
    al_dataloader_gathered_predictions,
    None,
    None,
]:
    assert not model.training
    for loader_batch in data_loader:
        state = call_hooks_stage_iterable(
            hook_iterable=batch_prep_hook,
            common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
            state=None,
        )
        batch = state["batch"]

        inputs = batch.inputs
        target_labels = batch.target_labels
        ids = batch.ids

        outputs = predict_on_batch(model=model, inputs=inputs)

        target_labels_copy = deepcopy(target_labels) if with_labels else None
        ids_copy = deepcopy(ids)

        yield outputs, target_labels_copy, ids_copy


def gather_data_loader_samples(
    data_loader: DataLoader,
    batch_prep_hook: Iterable[Callable],
    batch_prep_hook_kwargs: Dict[str, Any],
    n_samples: Union[int, None] = None,
) -> al_dataloader_gathered_raw:
    all_input_batches = []
    all_label_batches = []
    ids_total = []

    for loader_batch in data_loader:
        state = call_hooks_stage_iterable(
            hook_iterable=batch_prep_hook,
            common_kwargs={"loader_batch": loader_batch, **batch_prep_hook_kwargs},
            state=None,
        )
        batch = state["batch"]

        inputs: "al_input_batch" = batch.inputs
        target_labels: "al_training_labels_target" = batch.target_labels
        ids: "al_ids" = batch.ids

        all_input_batches.append(inputs)
        all_label_batches.append(target_labels)
        ids_total += [i for i in ids]

        if n_samples:
            if len(ids_total) >= n_samples:
                ids_total = ids_total[:n_samples]
                break

    all_input_batches_stacked = _stack_list_of_batch_dicts(
        list_of_batch_dicts=all_input_batches
    )
    all_target_label_batches_stacked = stack_list_of_output_target_dicts(
        list_of_target_batch_dicts=all_label_batches
    )

    if n_samples:
        inputs_final: "al_input_batch" = {}
        for input_name in all_input_batches_stacked.keys():
            input_subset = all_input_batches_stacked[input_name][:n_samples]
            inputs_final[input_name] = input_subset

        target_labels_final: "al_training_labels_target" = {}
        for output_name in all_target_label_batches_stacked.keys():
            target_labels_final[output_name] = {}

            cur_output = all_target_label_batches_stacked[output_name]
            for target_name in cur_output.keys():
                target_subset = cur_output[target_name][:n_samples]
                target_labels_final[output_name][target_name] = target_subset

    else:
        inputs_final, target_labels_final = (
            all_input_batches_stacked,
            all_target_label_batches_stacked,
        )

    return inputs_final, target_labels_final, ids_total


def stack_list_of_output_target_dicts(
    list_of_target_batch_dicts: List["al_training_labels_target"],
) -> "al_training_labels_target":
    """
    Spec:
        [batch_1, batch_2, batch_3]

        batch_1 =   {
                        'Output_Name_1': {'Target_Column_1': torch.Tensor(...)},
                        'Output_Name_2': {'Target_Column_2': torch.Tensor(...)},
                    }
                                                            with obs as rows in tensors
    """

    output_names = list_of_target_batch_dicts[0].keys()
    aggregated_batches: dict[str, dict[str, list[torch.Tensor]]] = {
        output_name: {} for output_name in output_names
    }
    for output_name in output_names:
        cur_output_targets = list_of_target_batch_dicts[0][output_name]
        for target_name in cur_output_targets.keys():
            aggregated_batches[output_name][target_name] = []

    for batch in list_of_target_batch_dicts:
        assert set(batch.keys()) == output_names

        for output_name in batch.keys():
            cur_output_batch = batch[output_name]
            for target_name in cur_output_batch.keys():
                cur_column_batch = cur_output_batch[target_name]
                cur_batch_value = [i for i in cur_column_batch]

                aggregated_batches[output_name][target_name] += cur_batch_value

    stacked_outputs: "al_training_labels_target" = {}
    for output_name, output_dict in aggregated_batches.items():
        cur_stacked_outputs = {
            key: _do_stack(list_of_elements=list_of_elements)
            for key, list_of_elements in output_dict.items()
        }
        stacked_outputs[output_name] = cur_stacked_outputs

    return stacked_outputs


def _stack_list_of_batch_dicts(
    list_of_batch_dicts: List["al_input_batch"],
) -> "al_input_batch":
    """
    Spec:
        [batch_1, batch_2, batch_3]
        batch_1 =   {
                        'input_1': torch.Tensor(...), # with obs as rows
                        'input_2': torch.Tensor(...),
                    }
    """

    target_columns = list_of_batch_dicts[0].keys()
    aggregated_batches: dict[str, list[torch.Tensor]] = {
        key: [] for key in target_columns
    }

    for batch in list_of_batch_dicts:
        assert set(batch.keys()) == target_columns

        for column in batch.keys():
            cur_column_batch = batch[column]
            aggregated_batches[column] += [i for i in cur_column_batch]

    stacked_inputs = {
        key: _do_stack(list_of_elements=list_of_elements)
        for key, list_of_elements in aggregated_batches.items()
    }

    return stacked_inputs


def _do_stack(
    list_of_elements: List[torch.Tensor],
) -> torch.Tensor:
    # check that they're all the same type
    list_types = set(type(i) for i in list_of_elements)
    assert len(list_types) == 1

    return torch.stack(list_of_elements)


def add_wd_to_model_params(
    model: nn.Module, wd: float
) -> List[Dict[str, nn.Parameter | float]]:
    """
    We want to skip adding weight decay to learnable activation parameters so as
    not to bias them towards 0.

    TODO:   Split this function in two, one to get the parameters and one to add the
            WD to them. Possibly we have to do it in-place here, not copy as we have
            tensors.

    Note: Since we are adding the weight decay manually here, the optimizer does not
    touch the parameter group weight decay at initialization.
    """
    _check_named_modules(model=model)

    param_list = []
    for name, param in model.named_parameters():
        cur_dict: dict[str, nn.Parameter | float]
        cur_dict = {"params": param}

        if "act_" in name:
            cur_dict["weight_decay"] = 0.0
        else:
            cur_dict["weight_decay"] = wd

        param_list.append(cur_dict)

    return param_list


def _check_named_modules(model: nn.Module):
    """
    We have this function as a safeguard to check that activations that have learnable
    parameters are named correctly (so that WD is not applied to them). Also, we want
    to make sure we don't have modules that are named 'incorrectly' and have the WD
    skipped when they should have it.
    """

    for name, module in model.named_modules():
        if name.startswith("act_"):
            assert isinstance(module, (Swish, nn.PReLU)), (name, module)

        if isinstance(module, (Swish, nn.PReLU)):
            assert "act_" in name, name


def run_lr_find(
    trainer_engine: Engine,
    train_dataloader: torch.utils.data.DataLoader,
    model: nn.Module,
    optimizer: Optimizer,
    output_folder: Path,
):
    lr_find_results = get_lr_range_results(
        trainer_engine=trainer_engine,
        train_dataloader=train_dataloader,
        model=model,
        optimizer=optimizer,
    )

    plot_lr_find_results(
        lr_values=lr_find_results.lr_history,
        loss_values=lr_find_results.loss_history,
        lr_suggestion=lr_find_results.lr_suggestion,
        output_folder=output_folder,
    )


@dataclass
class LRFindResults:
    lr_history_groups: list[list[float]]
    lr_history: list[float]
    loss_history: list[float]
    lr_suggestion: float


def get_lr_range_results(
    trainer_engine: Engine,
    train_dataloader: torch.utils.data.DataLoader,
    model: nn.Module,
    optimizer: Optimizer,
    num_iter: int = 500,
) -> LRFindResults:
    """
    We do a little hack with max_epochs and epoch_length because we don't pass that
    to the original trainer when running a normal training instance, which uses the
    default of max_epochs = 1. This is normally not a problem, as 1 epoch is
    generally more than for example 300 iterations, which should be enough for the LR
    test.

    However, in the cases where we have a small epoch length, the LR find will not
    run to completion as it only runs for one epoch. Hence, we need to patch the
    max_epochs here for the test.
    """
    lr_finder = FastaiLRFinder()

    def _extract_loss(x: dict[str, dict[str, dict[str, float]]]) -> float:
        return x["average"]["average"]["loss-average"]

    to_save = {"optimizer": optimizer, "model": model}
    with lr_finder.attach(
        trainer_engine,
        to_save=to_save,
        output_transform=_extract_loss,
        num_iter=num_iter,
        start_lr=1e-7,
    ) as trainer_with_lr_finder:
        logger.info("Running LR range test for max %d iterations.", num_iter)

        default_max_epochs = trainer_with_lr_finder.state.max_epochs
        trainer_with_lr_finder.state.max_epochs = 100
        trainer_with_lr_finder.state.epoch_length = len(train_dataloader)

        assert (
            trainer_with_lr_finder.state.max_epochs
            * trainer_with_lr_finder.state.epoch_length
        ) >= num_iter

        trainer_with_lr_finder.run(train_dataloader)
        trainer_with_lr_finder.state.max_epochs = default_max_epochs

        lr_results = deepcopy(lr_finder.get_results())
        lr_suggestions_all_groups = deepcopy(lr_finder.lr_suggestion())
        lr_suggestion_single = lr_suggestions_all_groups[0]

        logger.info(
            "LR Find: The model has %d parameter groups which can have different "
            "learning rates depending on implementation. For the LR find, "
            "we are using the LR defined in group 0, so make sure that is taken "
            "into consideration when analysing LR find results, the supplied LR "
            "for training might have to be scaled accordingly.",
            len(lr_results["lr"][0]),
        )

        return LRFindResults(
            lr_history_groups=lr_results["lr"],
            lr_history=[i[0] for i in lr_results["lr"]],
            loss_history=lr_results["loss"],
            lr_suggestion=lr_suggestion_single,
        )


def plot_lr_find_results(
    lr_values: list[float],
    loss_values: list[float],
    lr_suggestion: float,
    output_folder: Path,
) -> None:
    plt.plot(lr_values, loss_values)
    plt.xscale("log")
    plt.title(f"Learning Rate Search\nLR Suggestion: {lr_suggestion:.2e}")
    plt.xlabel("Learning Rate")
    plt.ylabel("Loss")
    plt.axvline(x=lr_suggestion, color="red", linewidth=1)

    plt.savefig(str(output_folder / "lr_search.pdf"))



from copy import copy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from random import sample
from typing import TYPE_CHECKING, Literal, Sequence, Union

from aislib.misc_utils import ensure_path_exists
from torch import nn
from torch.utils.data import DataLoader

from eir.data_load import label_setup
from eir.experiment_io.experiment_io import LoadedTrainExperiment
from eir.interpretation.interpretation import tabular_attribution_analysis_wrapper
from eir.predict_modules.predict_data import set_up_default_dataset
from eir.predict_modules.predict_input_setup import set_up_inputs_for_predict
from eir.predict_modules.predict_target_setup import (
    MergedPredictTargetLabels,
    get_target_labels_for_testing,
)
from eir.setup.config import Configs
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.output_setup import al_output_objects_as_dict
from eir.train import check_dataset_and_batch_size_compatibility
from eir.train_utils.step_logic import Hooks
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.predict import PredictExperiment

logger = get_logger(name=__name__)


def compute_predict_attributions(
    loaded_train_experiment: "LoadedTrainExperiment",
    predict_config: "PredictExperiment",
) -> None:
    gc = predict_config.configs.global_config

    background_source = (
        predict_config.predict_specific_cl_args.attribution_background_source
    )
    background_source_config = get_background_source_config(
        background_source_in_predict_cl_args=background_source,
        train_configs=loaded_train_experiment.configs,
        predict_configs=predict_config.configs,
    )
    background_dataloader = _get_predict_background_loader(
        batch_size=gc.be.batch_size,
        num_attribution_background_samples=gc.aa.attribution_background_samples,
        outputs_as_dict=loaded_train_experiment.outputs,
        configs=background_source_config,
        dataloader_workers=gc.be.dataloader_workers,
        loaded_hooks=loaded_train_experiment.hooks,
    )

    overloaded_train_experiment = _overload_train_experiment_for_predict_attributions(
        train_config=loaded_train_experiment,
        predict_config=predict_config,
    )

    attribution_output_folder_callable = partial(
        _get_predict_attribution_output_folder_target,
        predict_output_folder=Path(
            predict_config.predict_specific_cl_args.output_folder
        ),
    )

    tabular_attribution_analysis_wrapper(
        model=predict_config.model,
        experiment=overloaded_train_experiment,
        output_folder_target_callable=attribution_output_folder_callable,
        dataset_to_interpret=predict_config.test_dataset,
        background_loader=background_dataloader,
    )


def get_background_source_config(
    background_source_in_predict_cl_args: Literal["train", "predict"],
    train_configs: Configs,
    predict_configs: Configs,
) -> Configs:
    """
    TODO:   In the case of predict, make sure background and samples analysed are
            separated.
    """
    if background_source_in_predict_cl_args == "predict":
        logger.info(
            "Background for attribution analysis will be loaded from sources "
            "passed to predict.py."
        )
        return predict_configs

    elif background_source_in_predict_cl_args == "train":
        logger.info(
            "Background for attribution analysis will be loaded from sources "
            "previously used for training run with name '%s'.",
            train_configs.gc.be.output_folder,
        )
        return train_configs

    raise ValueError("Invalid background source. Expected 'train' or 'predict'.")


@dataclass
class LoadedTrainExperimentMixedWithPredict(LoadedTrainExperiment):
    model: nn.Module
    inputs: al_input_objects_as_dict


def _overload_train_experiment_for_predict_attributions(
    train_config: LoadedTrainExperiment,
    predict_config: "PredictExperiment",
) -> "LoadedTrainExperimentMixedWithPredict":
    """
    TODO:   Possibly set inputs=None as a field in LoadedTrainExperiment that then gets
            filled with test_inputs. When we do not need the weird monkey-patching here
            of the batch_prep_hooks, as the LoadedTrainExperiment will have the
            inputs attribute.
    """
    train_experiment_copy = copy(train_config)

    mixed_experiment_kwargs = train_experiment_copy.__dict__
    mixed_experiment_kwargs["model"] = predict_config.model
    mixed_experiment_kwargs["configs"] = predict_config.configs
    mixed_experiment_kwargs["inputs"] = predict_config.inputs

    mixed_experiment = LoadedTrainExperimentMixedWithPredict(**mixed_experiment_kwargs)

    return mixed_experiment


def _get_predict_background_loader(
    batch_size: int,
    num_attribution_background_samples: int,
    dataloader_workers: int,
    configs: Configs,
    outputs_as_dict: al_output_objects_as_dict,
    loaded_hooks: Union["Hooks", None],
):
    """
    TODO: Add option to choose whether to reuse train data as background,
          to use the data passed to the predict.py module, or possibly just
          an option to serialize the explainer from a training run and reuse
          that if passed as an option here.
    """

    background_ids_pool = label_setup.gather_all_ids_from_all_inputs(
        input_configs=configs.input_configs
    )

    custom_ops = None
    if loaded_hooks is not None:
        custom_ops = loaded_hooks.custom_column_label_parsing_ops

    target_labels = get_target_labels_for_testing(
        configs_overloaded_for_predict=configs,
        custom_column_label_parsing_ops=custom_ops,
        ids=background_ids_pool,
    )

    background_ids_sampled = _get_background_ids_sample(
        target_labels=target_labels,
        ids_inputs=background_ids_pool,
        num_attribution_background_samples=num_attribution_background_samples,
    )

    background_inputs_as_dict = set_up_inputs_for_predict(
        test_inputs_configs=configs.input_configs,
        ids=background_ids_sampled,
        hooks=loaded_hooks,
        output_folder=configs.gc.be.output_folder,
    )
    background_dataset = set_up_default_dataset(
        configs=configs,
        outputs_as_dict=outputs_as_dict,
        target_labels_dict=target_labels.label_dict,
        inputs_as_dict=background_inputs_as_dict,
        missing_ids_per_output=target_labels.missing_ids_per_output,
    )

    check_dataset_and_batch_size_compatibility(
        dataset=background_dataset,
        batch_size=batch_size,
        name="Test attribution background",
    )
    background_loader = DataLoader(
        dataset=background_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=dataloader_workers,
    )

    return background_loader


def _get_predict_attribution_output_folder_target(
    predict_output_folder: Path, output_name: str, column_name: str, input_name: str
) -> Path:
    attribution_output_folder = (
        predict_output_folder / output_name / column_name / "attributions" / input_name
    )
    ensure_path_exists(path=attribution_output_folder, is_folder=True)

    return attribution_output_folder


def _get_background_ids_sample(
    target_labels: MergedPredictTargetLabels,
    ids_inputs: Sequence[str],
    num_attribution_background_samples: int,
) -> list[str]:
    ids_target_set = set(target_labels.label_dict.keys())
    ids_inputs_set = set(ids_inputs)

    ids_common: list[str] = list(ids_target_set.intersection(ids_inputs_set))

    background_ids_sampled: list[str] = sample(
        population=ids_common,
        k=num_attribution_background_samples,
    )

    return background_ids_sampled



from pathlib import Path
from typing import Optional, Union

import numpy as np
import torch

from eir.data_load.data_augmentation import (
    make_random_omics_columns_missing,
    shuffle_random_omics_columns,
)
from eir.data_load.data_preparation_modules.common import _load_deeplake_sample
from eir.data_load.data_source_modules import deeplake_ops


def omics_load_wrapper(
    data_pointer: Union[Path, int],
    input_source: str,
    subset_indices: Optional[np.ndarray],
    deeplake_inner_key: Optional[str] = None,
) -> np.ndarray:
    if deeplake_ops.is_deeplake_dataset(data_source=input_source):
        assert deeplake_inner_key is not None
        assert isinstance(data_pointer, int)
        genotype_array_raw = _load_deeplake_sample(
            data_pointer=data_pointer,
            input_source=input_source,
            inner_key=deeplake_inner_key,
        )
    else:
        assert isinstance(data_pointer, Path)
        genotype_array_raw = np.load(str(data_pointer))

    if subset_indices is not None:
        genotype_array_raw = genotype_array_raw[:, subset_indices]

    genotype_array_raw_bool = genotype_array_raw.astype(bool)

    return genotype_array_raw_bool


def prepare_one_hot_omics_data(
    genotype_array: np.ndarray,
    na_augment_alpha: float,
    na_augment_beta: float,
    shuffle_augment_alpha: float,
    shuffle_augment_beta: float,
    test_mode: bool,
) -> torch.Tensor:
    """
    We use clone here to copy the original data, vs. using from_numpy
    which shares memory, causing us to modify the original data.
    """

    tensor_bool = torch.BoolTensor(genotype_array).unsqueeze(0).detach().clone()

    if not test_mode and na_augment_alpha > 0 and na_augment_beta > 0:
        tensor_bool = make_random_omics_columns_missing(
            omics_array=tensor_bool,
            na_augment_alpha=na_augment_alpha,
            na_augment_beta=na_augment_beta,
        )

    if not test_mode and shuffle_augment_alpha > 0 and shuffle_augment_beta > 0:
        tensor_bool = shuffle_random_omics_columns(
            omics_array=tensor_bool,
            shuffle_augment_alpha=shuffle_augment_alpha,
            shuffle_augment_beta=shuffle_augment_beta,
        )

    assert tensor_bool.dtype == torch.bool
    return tensor_bool



import math
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Tuple

import torch
import torch.nn as nn

from eir.models.fusion.fusion_attention import UniDirectionalCrossAttention
from eir.models.layers.cnn_layers import (
    ConvAttentionBlock,
    SEBlock,
    StochasticDepth,
    UpSamplingResidualBlock,
)
from eir.models.layers.norm_layers import GRN
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import FeatureExtractorInfo
    from eir.setup.input_setup_modules.common import DataDimensions


logger = get_logger(name=__name__)


@dataclass
class CNNUpscaleModelConfig:
    """
    :param channel_exp_base:
        Which power of 2 to use in order to set the number of channels in the network.
        For example, setting ``channel_exp_base=3`` means that 2**3=8 channels will be
        used.

    :param rb_do:
        Dropout in the convolutional residual blocks.

    :param stochastic_depth_p:
        Probability of dropping input.

    :param attention_inclusion_cutoff:
        If the dimension of width * height is less than this value, attention will be
        included in the model across channels and width * height as embedding
        dimension after that point
        (with the channels representing the length of the sequence).

    :param allow_pooling:
        Whether to allow adaptive average pooling in the model to match the target
        dimensions.

    :param num_ca_blocks:
          Number of cross-attention blocks to include in the model when fusing
          with other feature extractor outputs.

    :param up_every_n_blocks:
        If set, the model will upsample every n blocks. If not set, the model will
        use the down_every_n_blocks parameter in linked feature extractor, if
        that is the case and the pass-through upscale model is being used (
        the default for diffusion models, or if pass-through fusion model is used).
        Otherwise, will default to 2.

    :param n_final_extra_blocks:
        Number of extra blocks to add at the end of the model after upsampling
        has reached the target size.
    """

    channel_exp_base: int
    rb_do: float = 0.1
    stochastic_depth_p: float = 0.1
    attention_inclusion_cutoff: int = 256
    allow_pooling: bool = True
    num_ca_blocks: int = 1
    up_every_n_blocks: Optional[int] = None
    n_final_extra_blocks: int = 1


class CNNUpscaleResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        in_height: int,
        in_width: int,
        out_channels: int,
        stride: tuple[int, int],
        rb_do: float = 0.0,
        stochastic_depth_p: float = 0.0,
    ):
        super(CNNUpscaleResidualBlock, self).__init__()

        self.in_channels = in_channels
        self.in_height = in_height
        self.in_width = in_width

        self.stochastic_depth_p = stochastic_depth_p

        self.conv_ds = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=True,
            groups=in_channels,
        )

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=in_channels)

        self.conv_1 = nn.ConvTranspose2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=True,
        )

        self.rb_do = nn.Dropout2d(rb_do)
        self.act_1 = nn.GELU()

        self.grn = GRN(in_channels=out_channels)

        self.conv_2 = nn.Conv2d(
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )

        self.upsample_identity: nn.Module = nn.Identity()
        if in_channels != out_channels or stride != (1, 1):
            self.upsample_identity = nn.ConvTranspose2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=stride,
                padding=1,
                bias=True,
            )

        self.stochastic_depth = StochasticDepth(
            p=self.stochastic_depth_p,
            mode="batch",
        )

        self.se_block = SEBlock(
            channels=out_channels,
            reduction=16,
        )

    def forward(self, x: Any) -> Any:
        identity = self.upsample_identity(x)

        out = self.conv_ds(x)

        out = self.norm_1(out)

        out = self.conv_1(out)

        out = self.act_1(out)
        out = self.grn(out)
        out = self.rb_do(out)
        out = self.conv_2(out)

        channel_recalibrations = self.se_block(out)
        out = out * channel_recalibrations

        out = self.stochastic_depth(out)

        out = out + identity

        return out


def _do_add_attention(
    width: int,
    height: int,
    attention_inclusion_cutoff: int,
) -> bool:
    if attention_inclusion_cutoff == 0:
        return False

    if width * height <= attention_inclusion_cutoff:
        return True

    return False


def setup_blocks(
    in_channels: int,
    out_channels: int,
    initial_height: int,
    initial_width: int,
    target_height: int,
    target_width: int,
    attention_inclusion_cutoff: int,
    up_sample_every_n_blocks: int,
    n_final_extra_blocks: int,
    allow_pooling: bool = True,
) -> Tuple[nn.Sequential, int, int, int]:
    blocks = nn.Sequential()
    current_height = initial_height
    current_width = initial_width
    reduce_counter = 0
    block_counter = 0

    if _do_add_attention(
        width=current_width,
        height=current_height,
        attention_inclusion_cutoff=attention_inclusion_cutoff,
    ):
        cur_attention_block = ConvAttentionBlock(
            channels=in_channels,
            width=current_width,
            height=current_height,
        )
        blocks.add_module(
            name=f"block_{len(blocks)}",
            module=cur_attention_block,
        )

    while current_height < target_height or current_width < target_width:
        stride_height = 1
        stride_width = 1

        stride = (stride_height, stride_width)

        block_counter += 1
        if block_counter % 2 == 0 and reduce_counter < 4:
            out_channels = max(out_channels // 2, 1)
            reduce_counter += 1

        blocks.add_module(
            name=f"block_{len(blocks)}",
            module=CNNUpscaleResidualBlock(
                in_channels=in_channels,
                in_height=current_height,
                in_width=current_width,
                out_channels=out_channels,
                stride=stride,
            ),
        )

        in_channels = out_channels

        up_every = up_sample_every_n_blocks
        if up_every and block_counter % up_every == 0:

            do_height = current_height < target_height
            do_width = current_width < target_width

            up_sampling_block = UpSamplingResidualBlock(
                in_channels=in_channels,
                in_height=current_height,
                in_width=current_width,
                upsample_height=do_height,
                upsample_width=do_width,
            )
            blocks.add_module(
                name=f"block_{len(blocks)}_upsampling",
                module=up_sampling_block,
            )

            current_height = current_height * 2 if do_height else current_height
            current_width = current_width * 2 if do_width else current_width

        if _do_add_attention(
            width=current_width,
            height=current_height,
            attention_inclusion_cutoff=attention_inclusion_cutoff,
        ):
            cur_attention_block = ConvAttentionBlock(
                channels=out_channels,
                width=current_width,
                height=current_height,
            )
            blocks.add_module(
                name=f"block_{len(blocks)}",
                module=cur_attention_block,
            )

    for _ in range(n_final_extra_blocks):
        blocks.add_module(
            name=f"block_{len(blocks)}",
            module=CNNUpscaleResidualBlock(
                in_channels=in_channels,
                in_height=current_height,
                in_width=current_width,
                out_channels=out_channels,
                stride=(1, 1),
            ),
        )
        in_channels = out_channels

    not_matching = current_height != target_height or current_width != target_width
    if allow_pooling and not_matching:
        blocks.add_module(
            name="pooling",
            module=nn.AdaptiveAvgPool2d(
                output_size=(target_height, target_width),
            ),
        )
        current_height = target_height
        current_width = target_width

    return blocks, in_channels, current_height, current_width


class CNNUpscaleModel(nn.Module):
    def __init__(
        self,
        model_config: CNNUpscaleModelConfig,
        data_dimensions: "DataDimensions",
        target_dimensions: "DataDimensions",
    ):
        super(CNNUpscaleModel, self).__init__()

        self.model_config = model_config

        input_size = data_dimensions.num_elements()

        self.target_width = target_dimensions.width
        self.target_height = target_dimensions.height
        self.target_channels = target_dimensions.channels

        up_every_n_blocks = model_config.up_every_n_blocks
        if not up_every_n_blocks:
            logger.debug("Using default up_every_n_blocks=2.")
            up_every_n_blocks = 2

        ratio = math.sqrt(self.target_height * self.target_width / input_size)
        initial_height = int(self.target_height / ratio)
        initial_width = int(self.target_width / ratio)

        self.initial_layer = nn.Sequential(
            nn.Linear(
                in_features=input_size,
                out_features=initial_height * initial_width,
            ),
            nn.GELU(),
            nn.Unflatten(
                dim=1,
                unflattened_size=(1, initial_height, initial_width),
            ),
        )

        (
            self.blocks,
            self.block_channels,
            self.final_height,
            self.final_width,
        ) = setup_blocks(
            initial_height=initial_height,
            initial_width=initial_width,
            in_channels=1,
            target_height=self.target_height,
            target_width=self.target_width,
            out_channels=2**self.model_config.channel_exp_base,
            allow_pooling=self.model_config.allow_pooling,
            attention_inclusion_cutoff=self.model_config.attention_inclusion_cutoff,
            up_sample_every_n_blocks=up_every_n_blocks,
            n_final_extra_blocks=model_config.n_final_extra_blocks,
        )

        self.final_layer = nn.Conv2d(
            in_channels=self.block_channels,
            out_channels=self.target_channels,
            kernel_size=1,
        )

    @property
    def num_out_features(self) -> int:
        return self.target_channels * self.final_height * self.final_width

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.initial_layer(x)
        x = self.blocks(x)
        x = self.final_layer(x)
        return x


class CNNPassThroughUpscaleModel(nn.Module):
    def __init__(
        self,
        model_config: "CNNUpscaleModelConfig",
        feature_extractor_infos: dict[str, "FeatureExtractorInfo"],
        target_dimensions: "DataDimensions",
        output_name: str,
        diffusion_time_steps: Optional[int] = None,
    ) -> None:
        super(CNNPassThroughUpscaleModel, self).__init__()

        self.model_config = model_config
        self.feature_extractor_infos = feature_extractor_infos
        self.output_name = output_name

        if output_name not in feature_extractor_infos:
            raise ValueError(
                f"When using CNNPassThroughUpscaleModel, the output_name "
                f"'{output_name}' must be included as an input module, as the "
                f"passthrough model is intended to be linked with a feature "
                f"extractor."
            )

        cur_fei = feature_extractor_infos[output_name]
        cur_shape = cur_fei.output_shape
        assert cur_shape is not None

        up_every_n_blocks = model_config.up_every_n_blocks
        if not up_every_n_blocks:
            up_every_n_blocks = cur_fei.extras.get("down_every_n_blocks")
            if up_every_n_blocks:
                logger.debug(
                    "Using down_every_n_blocks=%s from linked feature extractor.",
                    up_every_n_blocks,
                )
            else:
                logger.debug("Using default up_every_n_blocks=2.")
                up_every_n_blocks = 2

        initial_channels = cur_shape[0]
        self.initial_height = cur_shape[1]
        self.initial_width = cur_shape[2]

        self.target_width = target_dimensions.width
        self.target_height = target_dimensions.height
        self.target_channels = target_dimensions.channels
        self.target_size = self.target_height * self.target_width
        self.diffusion_time_steps = diffusion_time_steps

        self.ca_layers: nn.ModuleDict = nn.ModuleDict()
        for name, fei in feature_extractor_infos.items():
            if name != output_name:
                cur_blocks = nn.ModuleList()

                for _ in range(self.model_config.num_ca_blocks):
                    cur_blocks.append(
                        CrossAttentionArrayOutBlock(
                            input_channels=initial_channels,
                            input_height=self.initial_height,
                            input_width=self.initial_width,
                            context_num_elements=fei.output_dimension,
                            context_dimension=fei.output_shape,
                        )
                    )

                self.ca_layers[name] = cur_blocks

        self.timestep_mixing_layer: nn.Identity | TimeStepMixingBlock = nn.Identity()
        if self.diffusion_time_steps is not None:
            self.timestep_mixing_layer = TimeStepMixingBlock(
                input_channels=initial_channels,
                input_height=self.initial_height,
                input_width=self.initial_width,
                n_time_steps=self.diffusion_time_steps,
            )

        (
            self.blocks,
            self.block_channels,
            self.final_height,
            self.final_width,
        ) = setup_blocks(
            in_channels=initial_channels,
            initial_height=self.initial_height,
            initial_width=self.initial_width,
            target_height=self.target_height,
            target_width=self.target_width,
            out_channels=2**self.model_config.channel_exp_base,
            allow_pooling=self.model_config.allow_pooling,
            attention_inclusion_cutoff=self.model_config.attention_inclusion_cutoff,
            up_sample_every_n_blocks=up_every_n_blocks,
            n_final_extra_blocks=model_config.n_final_extra_blocks,
        )

        self.final_layer = nn.Sequential(
            nn.Conv2d(
                in_channels=self.block_channels,
                out_channels=self.target_channels,
                kernel_size=1,
            ),
        )

    @property
    def num_out_features(self) -> int:
        return self.target_channels * self.final_height * self.final_width

    def timestep_embeddings(self, t: torch.Tensor) -> torch.Tensor:
        return self.timestep_mixing_layer.time_embedding(t)

    def forward(self, input: dict[str, torch.Tensor]) -> torch.Tensor:
        out = input[self.output_name]

        if len(out.shape) == 2:
            out = out.unflatten(
                dim=1,
                sizes=(1, self.initial_height, self.initial_width),
            )

        if self.diffusion_time_steps is not None:
            t_emb = input[f"__extras_{self.output_name}"]
            out = self.timestep_mixing_layer(input=out, t_emb=t_emb)

        for name, input_tensor in input.items():
            if name == self.output_name or name.startswith("__extras_"):
                continue

            cur_cross_attention: nn.ModuleList = self.ca_layers[name]
            for block in cur_cross_attention:
                out = block(input=out, context=input_tensor)

        out = self.blocks(out)
        out = self.final_layer(out)
        return out


class TimeStepMixingBlock(nn.Module):
    def __init__(
        self,
        input_channels: int,
        input_height: int,
        input_width: int,
        n_time_steps: int,
    ):
        super(TimeStepMixingBlock, self).__init__()

        self.input_channels = input_channels
        self.input_height = input_height
        self.input_width = input_width

        embedding_dim = input_height * input_width

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=input_channels)
        self.act_1 = nn.GELU()

        self.time_embedding = nn.Embedding(
            num_embeddings=n_time_steps,
            embedding_dim=embedding_dim,
        )

        self.cross_attention = UniDirectionalCrossAttention(
            dim=input_height * input_width,
            dim_head=input_height * input_width,
            context_dim=embedding_dim,
            heads=1,
        )

        self.grn = GRN(in_channels=input_channels)

    def forward(self, input: torch.Tensor, t_emb: torch.Tensor) -> torch.Tensor:
        identity = input

        out = self.norm_1(input)

        # (B, C, H, W) -> (B, C, H * W) -> (B, seq, emb_dim)
        out = out.view(
            out.shape[0],
            self.input_channels,
            self.input_height * self.input_width,
        )

        # (B, emb_dim) -> (B, 1, emb_dim)
        t_emb = t_emb.unsqueeze(1)

        out = self.cross_attention(x=out, context=t_emb)

        out = self.act_1(out)

        out = out.view(
            out.shape[0],
            self.input_channels,
            self.input_height,
            self.input_width,
        )

        out = self.grn(out)

        return out + identity


class CrossAttentionArrayOutBlock(nn.Module):
    def __init__(
        self,
        input_channels: int,
        input_height: int,
        input_width: int,
        context_num_elements: int,
        context_dimension: Optional[tuple[int, ...]],
    ):
        super(CrossAttentionArrayOutBlock, self).__init__()

        self.input_channels = input_channels
        self.input_height = input_height
        self.input_width = input_width
        self.context_dim = context_dimension

        self.conv_ds = nn.Conv2d(
            in_channels=input_channels,
            out_channels=input_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
            groups=input_channels,
        )

        self.norm_1 = nn.GroupNorm(num_groups=1, num_channels=input_channels)
        self.act_1 = nn.GELU()

        if self.context_dim is None:
            self.context_channels = 1
            self.context_height = 1
            self.context_width = context_num_elements
        else:
            self.context_channels = self.context_dim[0]
            self.context_height = self.context_dim[1]
            self.context_width = self.context_dim[2]

        context_emb_dim = self.context_height * self.context_width

        self.cross_attention = UniDirectionalCrossAttention(
            dim=input_height * input_width,
            dim_head=input_height * input_width,
            context_dim=context_emb_dim,
            heads=1,
        )

        self.conv_1 = nn.Conv2d(
            in_channels=input_channels,
            out_channels=input_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
        )

        self.grn = GRN(in_channels=input_channels)

    def forward(self, input: torch.Tensor, context: torch.Tensor) -> torch.Tensor:
        identity = input

        out = self.conv_ds(input)
        out = self.norm_1(out)

        # (B, C, H, W) -> (B, C, H * W) -> (B, seq, emb_dim)
        out = out.view(
            out.shape[0],
            self.input_channels,
            self.input_height * self.input_width,
        )

        # (B, C, H, W) -> (B, C, H * W) -> (B, seq, emb_dim)
        context = context.view(
            context.shape[0],
            self.context_channels,
            self.context_height * self.context_width,
        )

        out = self.cross_attention(x=out, context=context)
        out = self.act_1(out)

        out = out.view(
            out.shape[0],
            self.input_channels,
            self.input_height,
            self.input_width,
        )

        out = self.conv_1(out)
        out = self.grn(out)

        return out + identity



from functools import partial, update_wrapper
from typing import Any, Callable, Dict

import numpy as np
import torch
from PIL.Image import Image

from eir.data_load.data_preparation_modules.prepare_array import (
    array_load_wrapper,
    prepare_array_data,
)
from eir.data_load.data_preparation_modules.prepare_bytes import (
    bytes_load_wrapper,
    prepare_bytes_data,
)
from eir.data_load.data_preparation_modules.prepare_image import (
    image_load_wrapper,
    prepare_image_data,
)
from eir.data_load.data_preparation_modules.prepare_omics import (
    omics_load_wrapper,
    prepare_one_hot_omics_data,
)
from eir.data_load.data_preparation_modules.prepare_sequence import (
    prepare_sequence_data,
    sequence_load_wrapper,
)
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules.setup_array import ComputedArrayInputInfo
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_omics import ComputedOmicsInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.setup.schemas import (
    ByteInputDataConfig,
    ImageInputDataConfig,
    OmicsInputDataConfig,
    SequenceInputDataConfig,
    al_input_type_info,
)


def prepare_inputs_disk(
    inputs: Dict[str, Any], inputs_objects: "al_input_objects_as_dict", test_mode: bool
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}

    for input_name, data_pointer in inputs.items():
        input_object = inputs_objects[input_name]
        input_source = input_object.input_config.input_info.input_source
        input_type_info = input_object.input_config.input_type_info
        deeplake_inner_key = input_object.input_config.input_info.input_inner_key

        drop_rate = _get_modality_drop_rate(input_type_info=input_type_info)
        should_skip = _should_skip_modality(
            modality_dropout_rate=drop_rate,
            test_mode=test_mode,
        )

        if should_skip:
            continue

        match input_object:
            case ComputedOmicsInputInfo():
                array_raw = omics_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    deeplake_inner_key=deeplake_inner_key,
                    subset_indices=input_object.subset_indices,
                )

                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, OmicsInputDataConfig)
                input_prepared = prepare_one_hot_omics_data(
                    genotype_array=array_raw,
                    na_augment_alpha=input_type_info.na_augment_alpha,
                    na_augment_beta=input_type_info.na_augment_beta,
                    shuffle_augment_alpha=input_type_info.shuffle_augment_alpha,
                    shuffle_augment_beta=input_type_info.shuffle_augment_beta,
                    test_mode=test_mode,
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                sequence_tokenized = sequence_load_wrapper(
                    data_pointer=data_pointer,
                    input_source=input_source,
                    deeplake_inner_key=deeplake_inner_key,
                    split_on=input_type_info.split_on,
                    encode_func=input_object.encode_func,
                )
                input_prepared = prepare_sequence_data(
                    sequence_input_object=input_object,
                    cur_file_content_tokenized=sequence_tokenized,
                    test_mode=test_mode,
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)
                bytes_data = bytes_load_wrapper(
                    data_pointer=data_pointer,
                    dtype=input_type_info.byte_encoding,
                    input_source=input_source,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_bytes_data(
                    bytes_input_object=input_object,
                    bytes_data=bytes_data,
                    test_mode=test_mode,
                )

            case ComputedImageInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ImageInputDataConfig)
                image_data = image_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    image_mode=input_type_info.mode,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_image_data(
                    image_input_object=input_object,
                    image_data=image_data,
                    test_mode=test_mode,
                )

            case ComputedArrayInputInfo():
                array_data = array_load_wrapper(
                    input_source=input_source,
                    data_pointer=data_pointer,
                    deeplake_inner_key=deeplake_inner_key,
                )
                input_prepared = prepare_array_data(
                    array_data=array_data,
                    normalization_stats=input_object.normalization_stats,
                )

            case _:
                input_prepared = inputs[input_name]

        prepared_inputs[input_name] = input_prepared

    return prepared_inputs


def _get_modality_drop_rate(input_type_info: al_input_type_info) -> float:
    drop_rate = getattr(input_type_info, "modality_dropout_rate", 0.0)
    return drop_rate


def _should_skip_modality(modality_dropout_rate: float, test_mode: bool) -> bool:
    if test_mode or modality_dropout_rate == 0.0:
        return False

    should_skip = torch.rand(1) < modality_dropout_rate
    return bool(should_skip)


def prepare_inputs_memory(
    inputs: Dict[str, Any],
    inputs_objects: "al_input_objects_as_dict",
    test_mode: bool,
) -> Dict[str, torch.Tensor]:
    prepared_inputs = {}

    for name, data in inputs.items():
        input_object = inputs_objects[name]
        input_type_info = input_object.input_config.input_type_info

        drop_rate = _get_modality_drop_rate(input_type_info=input_type_info)
        should_skip = _should_skip_modality(
            modality_dropout_rate=drop_rate,
            test_mode=test_mode,
        )

        if should_skip:
            continue

        match input_object:
            case ComputedOmicsInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, OmicsInputDataConfig)
                input_prepared = prepare_one_hot_omics_data(
                    genotype_array=data,
                    na_augment_alpha=input_type_info.na_augment_alpha,
                    na_augment_beta=input_type_info.na_augment_beta,
                    shuffle_augment_alpha=input_type_info.shuffle_augment_alpha,
                    shuffle_augment_beta=input_type_info.shuffle_augment_beta,
                    test_mode=test_mode,
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                input_prepared = prepare_sequence_data(
                    sequence_input_object=input_object,
                    cur_file_content_tokenized=data,
                    test_mode=test_mode,
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)
                input_prepared = prepare_bytes_data(
                    bytes_input_object=input_object,
                    bytes_data=data,
                    test_mode=test_mode,
                )

            case ComputedImageInputInfo():
                input_prepared = prepare_image_data(
                    image_input_object=input_object,
                    image_data=data,
                    test_mode=test_mode,
                )

            case ComputedArrayInputInfo():
                input_prepared = prepare_array_data(
                    array_data=data,
                    normalization_stats=input_object.normalization_stats,
                )

            case _:
                input_prepared = inputs[name]

        prepared_inputs[name] = input_prepared

    return prepared_inputs


def typed_partial_for_hook(
    func: Callable[..., Any], *args: Any, **kwargs: Any
) -> Callable[..., np.ndarray | Image]:
    """
    Just to make mypy happy.
    """
    partial_func = partial(func, *args, **kwargs)
    return update_wrapper(partial_func, func)


def get_input_data_loading_hooks(
    inputs: al_input_objects_as_dict,
) -> dict[str, Callable[..., np.ndarray | Image]]:
    mapping = {}

    for input_name, input_object in inputs.items():
        common_kwargs = {
            "input_source": input_object.input_config.input_info.input_source,
            "deeplake_inner_key": input_object.input_config.input_info.input_inner_key,
        }

        match input_object:
            case ComputedOmicsInputInfo():
                mapping[input_name] = typed_partial_for_hook(
                    omics_load_wrapper,
                    **common_kwargs,
                    subset_indices=input_object.subset_indices,
                )

            case ComputedImageInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ImageInputDataConfig)
                mapping[input_name] = typed_partial_for_hook(
                    image_load_wrapper,
                    **common_kwargs,
                    image_mode=input_type_info.mode,
                )

            case ComputedSequenceInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, SequenceInputDataConfig)
                mapping[input_name] = typed_partial_for_hook(
                    sequence_load_wrapper,
                    **common_kwargs,
                    split_on=input_type_info.split_on,
                    encode_func=input_object.encode_func,
                )

            case ComputedBytesInputInfo():
                input_type_info = input_object.input_config.input_type_info
                assert isinstance(input_type_info, ByteInputDataConfig)
                mapping[input_name] = typed_partial_for_hook(
                    bytes_load_wrapper,
                    **common_kwargs,
                    dtype=input_type_info.byte_encoding,
                )

            case ComputedArrayInputInfo():
                mapping[input_name] = typed_partial_for_hook(
                    array_load_wrapper,
                    **common_kwargs,
                )

    return mapping



import csv
import warnings
from copy import copy
from dataclasses import dataclass
from functools import partial, wraps
from pathlib import Path
from statistics import mean
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import ensure_path_exists
from scipy.special import softmax
from scipy.stats import pearsonr
from sklearn.metrics import (
    accuracy_score,
    average_precision_score,
    matthews_corrcoef,
    r2_score,
    roc_auc_score,
)
from sklearn.preprocessing import StandardScaler, label_binarize
from torch import nn
from torch.linalg import vector_norm

from eir.data_load.data_utils import get_output_info_generator
from eir.setup.schema_modules.output_schemas_tabular import TabularOutputTypeConfig
from eir.setup.schemas import OutputConfig
from eir.target_setup.target_label_setup import MissingTargetsInfo
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.data_load.label_setup import (  # noqa: F401
        al_label_transformers,
        al_label_transformers_object,
        al_target_columns,
    )
    from eir.models.input.omics.omics_models import al_omics_models  # noqa: F401
    from eir.models.meta.meta_utils import FeatureExtractorProtocolWithL1
    from eir.setup.output_setup import al_output_objects_as_dict
    from eir.train import Experiment, al_criteria_dict  # noqa: F401
    from eir.train_utils.step_logic import al_training_labels_target
    from eir.train_utils.train_handlers import HandlerConfig

# aliases
# output_name -> target_name -> metric name: value
al_step_metric_dict = Dict[str, Dict[str, Dict[str, float]]]

logger = get_logger(name=__name__)


class MetricFunctionProtocol(Protocol):
    def __call__(
        self,
        outputs: np.ndarray,
        labels: np.ndarray,
        column_name: str,
        output_name: str,
    ) -> float: ...


class AverageMetricFunctionProtocol(Protocol):
    def __call__(
        self,
        metric_dict: al_step_metric_dict,
        output_name: str,
        column_name: str,
    ) -> float: ...


al_averaging_functions_dict = Dict[str, AverageMetricFunctionProtocol]
al_metric_record_dict = Dict[
    str, Tuple["MetricRecord", ...] | "al_averaging_functions_dict"
]

al_cat_averaging_metric_choices = Sequence[
    Literal["mcc", "acc", "roc-auc-macro", "ap-macro"]
]
al_con_averaging_metric_choices = Sequence[Literal["r2", "pcc", "loss"]]


@dataclass()
class MetricRecord:
    name: str
    function: MetricFunctionProtocol
    only_val: bool = False
    minimize_goal: bool = False


def calculate_batch_metrics(
    outputs_as_dict: "al_output_objects_as_dict",
    outputs: Dict[str, Dict[str, torch.Tensor]],
    labels: "al_training_labels_target",
    mode: str,
    metric_record_dict: al_metric_record_dict,
) -> al_step_metric_dict:
    assert mode in ["val", "train"]

    target_columns_gen = get_output_info_generator(outputs_as_dict=outputs_as_dict)

    master_metric_dict: al_step_metric_dict = {}

    for output_name, output_target_type, target_name in target_columns_gen:
        cur_metric_dict: dict[str, float] = {}

        if output_name not in master_metric_dict:
            master_metric_dict[output_name] = {}

        if output_target_type == "general":
            master_metric_dict[output_name][target_name] = cur_metric_dict

        elif output_target_type in ("con", "cat"):
            cur_output_object = outputs_as_dict[output_name]
            cur_output_type = cur_output_object.output_config.output_info.output_type
            assert cur_output_type == "tabular"

            al_record = Tuple[MetricRecord, ...]
            cur_record = metric_record_dict[output_target_type]
            assert isinstance(cur_record, tuple)
            cur_metric_records: al_record = cur_record

            cur_outputs = outputs[output_name][target_name]
            cur_outputs_np = cur_outputs.detach().cpu().to(dtype=torch.float32).numpy()

            cur_labels = labels[output_name][target_name]
            cur_labels_np = cur_labels.cpu().to(dtype=torch.float32).numpy()

            for metric_record in cur_metric_records:
                if metric_record.only_val and mode == "train":
                    continue

                cur_key = f"{output_name}_{target_name}_{metric_record.name}"
                cur_value = metric_record.function(
                    outputs=cur_outputs_np,
                    labels=cur_labels_np,
                    column_name=target_name,
                    output_name=output_name,
                )

                cur_metric_dict[cur_key] = cur_value

            master_metric_dict[output_name][target_name] = cur_metric_dict
        else:
            raise NotImplementedError()

    return master_metric_dict


def add_loss_to_metrics(
    outputs_as_dict: "al_output_objects_as_dict",
    losses: Dict[str, Dict[str, torch.Tensor]],
    metric_dict: al_step_metric_dict,
) -> al_step_metric_dict:
    target_columns_gen = get_output_info_generator(outputs_as_dict=outputs_as_dict)
    metric_dict_copy = copy(metric_dict)

    for output_name, column_type, column_name in target_columns_gen:
        cur_metric_dict = metric_dict_copy[output_name][column_name]
        cur_key = f"{output_name}_{column_name}_loss"
        cur_metric_dict[cur_key] = losses[output_name][column_name].item()

    return metric_dict_copy


def add_multi_task_average_metrics(
    batch_metrics_dict: al_step_metric_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    loss: float,
    performance_average_functions: Optional["al_averaging_functions_dict"],
) -> al_step_metric_dict:
    average = {
        "average": {
            "loss-average": loss,
        }
    }

    if performance_average_functions is not None:
        average_performance = average_performances_across_tasks(
            metric_dict=batch_metrics_dict,
            outputs_as_dict=outputs_as_dict,
            performance_calculation_functions=performance_average_functions,
        )
        average["average"]["perf-average"] = average_performance

    batch_metrics_dict["average"] = average

    return batch_metrics_dict


def average_performances_across_tasks(
    metric_dict: al_step_metric_dict,
    outputs_as_dict: "al_output_objects_as_dict",
    performance_calculation_functions: "al_averaging_functions_dict",
) -> float:
    target_columns_gen = get_output_info_generator(
        outputs_as_dict=outputs_as_dict,
    )

    all_metrics = []

    for output_name, column_type, column_name in target_columns_gen:
        cur_metric_func = performance_calculation_functions[column_type]

        cur_value = cur_metric_func(
            metric_dict=metric_dict, output_name=output_name, column_name=column_name
        )

        if not np.isnan(cur_value):
            all_metrics.append(cur_value)

    average = np.array(all_metrics).mean()

    return average


def handle_empty(default_value: float, metric_name: Optional[str] = None):
    """
    This can happen when modelling on multiple outputs, which can vary in their
    sparsity, and by chance some outputs are empty in a batch.
    """

    def decorator(func):
        logged = False

        @wraps(func)
        def wrapper(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs):
            nonlocal logged
            if (outputs.size == 0 or labels.size == 0) and not logged:
                logger.info(
                    f"Empty inputs encountered in "
                    f"{metric_name or func.__name__}, "
                    f"returning default value: {default_value}"
                )
                logged = True
            if outputs.size == 0 or labels.size == 0:
                return default_value
            return func(outputs, labels, *args, **kwargs)

        return wrapper

    return decorator


def handle_class_mismatch(default_value: float, metric_name: Optional[str] = None):
    """
    Decorator to handle cases where the number of unique classes in 'labels'
    does not match the number of columns in 'outputs'. This scenario can occur
    in multi-class classification tasks with sparse or imbalanced outputs.
    """

    def decorator(func):
        logged = False

        @wraps(func)
        def wrapper(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs):
            nonlocal logged
            unique_classes = np.unique(labels)
            # For binary classification we have this special case
            if outputs.shape[1] == 2 and len(unique_classes) == 1:
                pass
            elif len(unique_classes) != outputs.shape[1]:
                if not logged:
                    logger.info(
                        f"Class mismatch encountered in "
                        f"{metric_name or func.__name__}, "
                        f"expected number of classes: {outputs.shape[1]}, "
                        f"found unique classes in labels: {len(unique_classes)}. "
                        f"Returning default value: {default_value}"
                    )
                    logged = True
                return default_value
            return func(outputs, labels, *args, **kwargs)

        return wrapper

    return decorator


@handle_empty(default_value=np.nan, metric_name="MCC")
def calc_mcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    prediction = np.argmax(a=outputs, axis=1)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)
        mcc = matthews_corrcoef(labels, prediction)

    return mcc


@handle_class_mismatch(default_value=np.nan, metric_name="ROC-AUC")
@handle_empty(default_value=np.nan, metric_name="ROC-AUC")
def calc_roc_auc_ovo(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    """
    TODO:   In rare scenarios, we might run into the issue of not having all labels
            represented in the labels array (i.e. labels were in train, but not in
            valid). This is not a problem for metrics like MCC / accuracy, but we
            will have to account for this here and in the AP calculation, possibly
            by ignoring columns in outputs and label_binarize outputs where the columns
            returned from label_binarize are all 0.
    """

    average = "macro"

    if outputs.shape[1] > 2:
        outputs = softmax(x=outputs, axis=1)
    else:
        outputs = outputs[:, 1]

    roc_auc = roc_auc_score(
        y_true=labels,
        y_score=outputs,
        average=average,
        multi_class="ovo",
    )
    return roc_auc


@handle_class_mismatch(default_value=np.nan, metric_name="AP")
@handle_empty(default_value=np.nan, metric_name="AP")
def calc_average_precision(
    outputs: np.ndarray, labels: np.ndarray, *args, **kwargs
) -> float:
    average = "macro"

    if outputs.shape[1] > 2:
        labels = label_binarize(y=labels, classes=sorted(np.unique(labels)))
    else:
        outputs = outputs[:, 1]

    average_precision = average_precision_score(
        y_true=labels,
        y_score=outputs,
        average=average,
    )

    return average_precision


@handle_empty(default_value=np.nan, metric_name="ACC")
def calc_acc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    pred = np.argmax(outputs, axis=1)

    accuracy = accuracy_score(y_true=labels, y_pred=pred)
    return accuracy


@handle_empty(default_value=np.nan, metric_name="PCC")
def calc_pcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    if len(outputs) < 2:
        return 0.0

    pcc = pearsonr(x=labels.squeeze(), y=outputs.squeeze())[0]
    return pcc


@handle_empty(default_value=np.nan, metric_name="R2")
def calc_r2(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    if len(outputs) < 2:
        return 0.0

    r2 = r2_score(y_true=labels.squeeze(), y_pred=outputs.squeeze())
    return r2


@handle_empty(default_value=np.nan, metric_name="RMSE")
def calc_rmse(
    outputs: np.ndarray,
    labels: np.ndarray,
    target_transformers: Dict[str, Dict[str, StandardScaler]],
    output_name: str,
    column_name: str,
    *args,
    **kwargs,
) -> float:
    cur_target_transformer = target_transformers[output_name][column_name]

    labels_2d = labels.reshape(-1, 1)
    outputs_2d = outputs.reshape(-1, 1)

    mean_ = cur_target_transformer.mean_
    scale_ = cur_target_transformer.scale_

    labels = (labels_2d * scale_ + mean_).squeeze()
    predictions = (outputs_2d * scale_ + mean_).squeeze()

    if np.shape(labels):
        rmse = np.sqrt(np.mean((labels - predictions) ** 2))
    else:
        rmse = np.sqrt((labels - predictions) ** 2)

    return rmse


class LogEmptyLossProtocol(Protocol):
    def __call__(self, output_name: str, output_head_name: str) -> None: ...


def log_empty_loss_once() -> LogEmptyLossProtocol:
    logged_combinations = set()

    def log(output_name: str, output_head_name: str) -> None:
        if (output_name, output_head_name) not in logged_combinations:
            logger.info(
                f"Empty output batch encountered for {output_name},"
                f" {output_head_name}; "
                f"setting loss to NaN for this batch and future empty batches. "
                f"Empty batches will not be used for training."
            )
            logged_combinations.add((output_name, output_head_name))

    return log


def calculate_prediction_losses(
    criteria: "al_criteria_dict",
    inputs: dict[str, dict[str, torch.Tensor]],
    targets: dict[str, dict[str, torch.Tensor]],
    log_empty_loss_callable: LogEmptyLossProtocol,
) -> dict[str, dict[str, torch.Tensor]]:
    """
    We check for empty output tensors and log a warning if we encounter them.
    This can happen when modelling on multiple outputs, which can vary in their
    sparsity, and by chance some outputs are empty in a batch.
    """
    losses_dict: dict[str, dict[str, torch.Tensor]] = {}

    for output_name, target_dict in targets.items():
        for output_head_name, cur_inner_target in target_dict.items():
            if output_head_name in inputs[output_name]:
                input_tensor = inputs[output_name][output_head_name]
                criterion = criteria[output_name][output_head_name]

                if output_name not in losses_dict:
                    losses_dict[output_name] = {}

                if input_tensor.nelement() > 0 and cur_inner_target.nelement() > 0:
                    computed_loss = criterion(
                        input=input_tensor,
                        target=cur_inner_target,
                    )
                else:
                    log_empty_loss_callable(
                        output_name=output_name, output_head_name=output_head_name
                    )
                    computed_loss = torch.tensor(np.nan, requires_grad=True)

                losses_dict[output_name][output_head_name] = computed_loss
            else:
                raise KeyError(f"Missing input for {output_name} {output_head_name}")

    return losses_dict


def aggregate_losses(losses_dict: dict[str, dict[str, torch.Tensor]]) -> torch.Tensor:
    losses_values = []
    for output_name, targets_for_output_dict in losses_dict.items():
        for loss in targets_for_output_dict.values():
            if not torch.isnan(loss).any():
                losses_values.append(loss)

    if not losses_values:
        return torch.tensor(np.nan, requires_grad=True)

    average_loss = torch.mean(torch.stack(losses_values))
    return average_loss


def get_uncertainty_loss_hook(
    output_configs: Sequence[OutputConfig],
    device: str,
) -> Callable:
    uncertainty_loss_modules = {}
    for output_config in output_configs:
        if output_config.output_info.output_type != "tabular":
            continue

        output_type_info = output_config.output_type_info
        assert isinstance(output_type_info, TabularOutputTypeConfig)

        if not output_type_info.uncertainty_weighted_mt_loss:
            continue

        logger.info(
            f"Adding uncertainty loss for {output_config.output_info.output_name}."
        )

        cur_target_cat_columns = list(output_type_info.target_cat_columns)
        cur_target_con_columns = list(output_type_info.target_con_columns)
        loss_module = UncertaintyMultiTaskLoss(
            target_cat_columns=cur_target_cat_columns,
            target_con_columns=cur_target_con_columns,
            device=device,
        )
        uncertainty_loss_modules[output_config.output_info.output_name] = loss_module

    if len(uncertainty_loss_modules) == 0:
        raise ValueError("Expected at least one uncertainty loss module.")

    hook = partial(
        hook_add_uncertainty_loss, uncertainty_modules=uncertainty_loss_modules
    )

    return hook


def hook_add_uncertainty_loss(
    state: Dict,
    uncertainty_modules: Dict[str, "UncertaintyMultiTaskLoss"],
    loss_key: str = "per_target_train_losses",
    *args,
    **kwargs,
) -> Dict[str, Dict[str, torch.Tensor]]:
    """
    Note that we only update the relevant losses in the base dict.
    """

    base_losses_dict = state[loss_key]
    updated_losses = copy(base_losses_dict)

    for output_name, module in uncertainty_modules.items():
        cur_module = uncertainty_modules[output_name]
        cur_loss_dict = base_losses_dict[output_name]
        cur_uncertainty_losses = cur_module(losses_dict=cur_loss_dict)
        updated_losses[output_name] = cur_uncertainty_losses

    state_updates = {loss_key: updated_losses}

    return state_updates


class UncertaintyMultiTaskLoss(nn.Module):
    def __init__(
        self,
        target_cat_columns: List[str],
        target_con_columns: List[str],
        device: str,
    ):
        super().__init__()

        self.target_cat_columns = target_cat_columns
        self.target_con_columns = target_con_columns
        self.device = device

        self.log_vars = self._construct_params(
            cur_target_columns=self.target_cat_columns + self.target_con_columns,
            device=self.device,
        )

    @staticmethod
    def _construct_params(
        cur_target_columns: List[str], device: str
    ) -> dict[str, torch.Tensor]:
        param_dict: dict[str, torch.Tensor] = {}
        for column_name in cur_target_columns:
            cur_param = nn.Parameter(torch.zeros(1), requires_grad=True).to(
                device=device
            )
            param_dict[column_name] = cur_param

        return param_dict

    def _calc_uncertainty_loss(
        self, name: str, loss_value: torch.Tensor
    ) -> torch.Tensor:
        log_var = self.log_vars[name]
        scalar = 2.0 if name in self.target_cat_columns else 1.0

        precision = torch.exp(-log_var)
        loss = scalar * torch.sum(precision * loss_value + log_var)

        return loss

    def forward(self, losses_dict: Dict) -> Dict[str, torch.Tensor]:
        losses_uncertain = {}

        for target_column, loss_value_base in losses_dict.items():
            loss_value_uncertain = self._calc_uncertainty_loss(
                name=target_column, loss_value=loss_value_base
            )
            losses_uncertain[target_column] = loss_value_uncertain

        return losses_uncertain


def hook_add_l1_loss(
    experiment: "Experiment",
    state: Dict,
    loss_key: str = "loss",
    *args,
    **kwargs,
) -> Dict:
    """
    TODO: Do the validation outside of the actual hook.
    """
    model_configs = experiment.inputs

    l1_loss = torch.tensor(0.0, device=experiment.configs.gc.be.device)
    for input_name, input_module in experiment.model.input_modules.items():
        cur_model_config = model_configs[input_name].input_config.model_config
        cur_model_init_config = cur_model_config.model_init_config

        current_l1 = getattr(cur_model_init_config, "l1", None)
        has_l1_weights = hasattr(input_module, "l1_penalized_weights")

        if current_l1 and not has_l1_weights:
            raise AttributeError(
                f"Module {input_module} for input name {input_name} does not have"
                f"l1_penalized_weights attribute."
            )

        if has_l1_weights and current_l1:
            input_module_with_l1 = cast("FeatureExtractorProtocolWithL1", input_module)
            cur_l1_loss = get_model_l1_loss(
                model=input_module_with_l1, l1_weight=current_l1
            )
            l1_loss += cur_l1_loss

    updated_loss = state[loss_key] + l1_loss

    state_updates = {loss_key: updated_loss}

    return state_updates


def get_model_l1_loss(
    model: "FeatureExtractorProtocolWithL1", l1_weight: float
) -> torch.Tensor:
    l1_loss = calc_l1_loss(
        weight_tensor=model.l1_penalized_weights, l1_weight=l1_weight
    )
    return l1_loss


def calc_l1_loss(weight_tensor: torch.Tensor, l1_weight: float):
    l1_loss = vector_norm(weight_tensor, ord=1, dim=None) * l1_weight
    return l1_loss


def persist_metrics(
    handler_config: "HandlerConfig",
    metrics_dict: "al_step_metric_dict",
    iteration: int,
    write_header: bool,
    prefixes: Dict[str, str],
    writer_funcs: Union[None, Dict[str, Dict[str, Callable]]] = None,
) -> None:

    hc = handler_config
    exp = handler_config.experiment
    gc = exp.configs.global_config

    target_generator = get_output_info_generator(outputs_as_dict=exp.outputs)

    metrics_files = get_metrics_files(
        target_generator=target_generator,
        run_folder=hc.run_folder,
        train_or_val_target_prefix=f"{prefixes['metrics']}",
        detail_level=gc.evaluation_checkpoint.saved_result_detail_level,
    )

    if write_header:
        _ensure_metrics_paths_exists(metrics_files=metrics_files)

    for output_name, target_and_file_dict in metrics_files.items():
        for target_name, target_history_file in target_and_file_dict.items():
            cur_metric_dict = metrics_dict[output_name][target_name]

            if writer_funcs:
                cur_func = writer_funcs[output_name][target_name]
            else:
                cur_func = get_buffered_metrics_writer(buffer_interval=1)

            cur_func(
                filepath=target_history_file,
                metrics=cur_metric_dict,
                iteration=iteration,
                write_header=write_header,
            )


def get_metrics_files(
    target_generator: Generator[Tuple[str, str, str], None, None],
    run_folder: Path,
    train_or_val_target_prefix: str,
    detail_level: int,
) -> dict[str, dict[str, Path]]:
    assert train_or_val_target_prefix in ["validation_", "train_"]

    path_dict: dict[str, dict[str, Path]] = {}
    for output_name, column_type, target_column in target_generator:
        if detail_level <= 1:
            continue

        if output_name not in path_dict:
            path_dict[output_name] = {}

        cur_file_name = train_or_val_target_prefix + target_column + "_history.log"
        cur_path = Path(
            run_folder,
            "results",
            output_name,
            target_column,
            cur_file_name,
        )
        path_dict[output_name][target_column] = cur_path

    average_loss_training_metrics_file = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix=train_or_val_target_prefix
    )
    path_dict["average"] = {"average": average_loss_training_metrics_file}

    return path_dict


def get_average_history_filepath(
    run_folder: Path, train_or_val_target_prefix: str
) -> Path:
    assert train_or_val_target_prefix in ["validation_", "train_"]
    metrics_file_path = run_folder / f"{train_or_val_target_prefix}average_history.log"
    return metrics_file_path


def _ensure_metrics_paths_exists(metrics_files: Dict[str, Dict[str, Path]]) -> None:
    for output_name, target_and_file_dict in metrics_files.items():
        for path in target_and_file_dict.values():
            ensure_path_exists(path=path)


def get_buffered_metrics_writer(buffer_interval: int):
    buffer = []

    def append_metrics_to_file(
        filepath: Path,
        metrics: Dict[str, float],
        iteration: int,
        write_header=False,
    ) -> None:
        nonlocal buffer

        dict_to_write = {**{"iteration": iteration}, **metrics}

        if write_header:
            with open(str(filepath), "a") as logfile:
                fieldnames = ["iteration"] + sorted(metrics.keys())
                writer = csv.DictWriter(logfile, fieldnames=fieldnames)

                if write_header:
                    writer.writeheader()

        if iteration % buffer_interval == 0:
            buffer.append(dict_to_write)

            with open(str(filepath), "a") as logfile:
                fieldnames = ["iteration"] + sorted(metrics.keys())
                writer = csv.DictWriter(logfile, fieldnames=fieldnames)

                source = buffer if buffer_interval != 1 else [dict_to_write]
                for row in source:
                    writer.writerow(row)

            buffer = []

        else:
            buffer.append(dict_to_write)

    return append_metrics_to_file


def read_metrics_history_file(file_path: Path) -> pd.DataFrame:
    df = pd.read_csv(file_path, index_col="iteration")

    return df


def get_metrics_dataframes(
    results_dir: Path, target_string: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    train_history_path = read_metrics_history_file(
        results_dir / f"train_{target_string}_history.log"
    )
    valid_history_path = read_metrics_history_file(
        results_dir / f"validation_{target_string}_history.log"
    )

    return train_history_path, valid_history_path


def get_default_metrics(
    target_transformers: Dict[str, "al_label_transformers"],
    cat_averaging_metrics: Optional[al_cat_averaging_metric_choices],
    con_averaging_metrics: Optional[al_con_averaging_metric_choices],
) -> "al_metric_record_dict":
    mcc = MetricRecord(
        name="mcc",
        function=calc_mcc,
    )
    acc = MetricRecord(
        name="acc",
        function=calc_acc,
    )
    roc_auc_macro = MetricRecord(
        name="roc-auc-macro",
        function=calc_roc_auc_ovo,
        only_val=True,
    )
    ap_macro = MetricRecord(
        name="ap-macro",
        function=calc_average_precision,
        only_val=True,
    )

    rmse = MetricRecord(
        name="rmse",
        function=partial(calc_rmse, target_transformers=target_transformers),
        minimize_goal=True,
    )
    r2 = MetricRecord(
        name="r2",
        function=calc_r2,
        only_val=True,
    )
    pcc = MetricRecord(
        name="pcc",
        function=calc_pcc,
        only_val=True,
    )

    cat_for_avg, con_for_avg = parse_averaging_metrics(
        cat_averaging_metrics=cat_averaging_metrics,
        con_averaging_metrics=con_averaging_metrics,
    )
    averaging_functions = get_performance_averaging_functions(
        cat_metric_names=cat_for_avg,
        con_metric_names=con_for_avg,
    )

    default_metrics: al_metric_record_dict = {
        "cat": (mcc, acc, roc_auc_macro, ap_macro),
        "con": (rmse, r2, pcc),
        "averaging_functions": averaging_functions,
    }
    return default_metrics


def parse_averaging_metrics(
    cat_averaging_metrics: Optional[al_cat_averaging_metric_choices],
    con_averaging_metrics: Optional[al_con_averaging_metric_choices],
) -> tuple[al_cat_averaging_metric_choices, al_con_averaging_metric_choices]:
    cat_parsed, con_parsed = _get_default_averaging_metrics()

    if cat_averaging_metrics:
        _validate_metrics(
            passed_in_metrics=cat_averaging_metrics,
            expected_metrics=[
                "loss",
                "acc",
                "mcc",
                "roc-auc-macro",
                "ap-macro",
            ],
            target_type="categorical",
        )
        cat_parsed = cat_averaging_metrics
    if con_averaging_metrics:
        _validate_metrics(
            passed_in_metrics=con_averaging_metrics,
            expected_metrics=[
                "loss",
                "rmse",
                "pcc",
                "r2",
            ],
            target_type="continuous",
        )
        con_parsed = con_averaging_metrics

    return cat_parsed, con_parsed


def _validate_metrics(
    passed_in_metrics: Sequence[str], expected_metrics: Sequence[str], target_type: str
) -> None:
    for metric in passed_in_metrics:
        if metric not in expected_metrics:
            raise ValueError(
                f"Metric {metric} not found in expected metrics {expected_metrics} for "
                f" {target_type} targets."
            )


def _get_default_averaging_metrics() -> (
    tuple[al_cat_averaging_metric_choices, al_con_averaging_metric_choices]
):

    cat_names: list[Literal["mcc", "roc-auc-macro", "ap-macro", "acc"]]
    con_names: list[Literal["loss", "pcc", "r2"]]

    cat_names = ["mcc", "roc-auc-macro", "ap-macro"]
    con_names = ["loss", "pcc", "r2"]

    return cat_names, con_names


def get_performance_averaging_functions(
    cat_metric_names: al_cat_averaging_metric_choices,
    con_metric_names: al_con_averaging_metric_choices,
) -> al_averaging_functions_dict:
    """
    Note that we have the mean(values) else 0.0 to account for some values not being
    computed on the training batches, e.g. ROC-AUC, due some metrics possibly
    raising errors if e.g. there are only negative labels in a batch.
    """

    logger.info(
        "Tabular output performance averaging functions across tasks set to averages "
        "of %s for categorical targets and %s for continuous targets. These "
        "values are used to determine overall performance (using the validation set), "
        "which is used to control factors such as early stopping and LR scheduling. "
        "Other output cases use 1.0-LOSS by default.",
        [i.upper() for i in cat_metric_names],
        [i.upper().replace("LOSS", "1.0-LOSS") for i in con_metric_names],
    )

    def _calc_cat_averaging_value(
        metric_dict: "al_step_metric_dict",
        output_name: str,
        column_name: str,
        metric_names: al_cat_averaging_metric_choices,
    ) -> float:
        values = []
        for metric_name in metric_names:
            combined_key = f"{output_name}_{column_name}_{metric_name}"
            value = metric_dict[output_name][column_name].get(combined_key, None)

            if value is None:
                continue

            values.append(value)

        return mean(values) if values else 0.0

    def _calc_con_averaging_value(
        metric_dict: "al_step_metric_dict",
        output_name: str,
        column_name: str,
        metric_names: al_con_averaging_metric_choices,
    ) -> float:
        values = []
        for metric_name in metric_names:
            combined_key = f"{output_name}_{column_name}_{metric_name}"
            value = metric_dict[output_name][column_name].get(combined_key, None)

            if value is None:
                continue

            if metric_name == "loss":
                value = 1.0 - value

            values.append(value)

        return mean(values) if values else 0.0

    performance_averaging_functions: al_averaging_functions_dict = {
        "cat": partial(_calc_cat_averaging_value, metric_names=cat_metric_names),
        "con": partial(_calc_con_averaging_value, metric_names=con_metric_names),
        "general": partial(_calc_con_averaging_value, metric_names=["loss"]),
    }

    return performance_averaging_functions


@dataclass
class FilteredOutputsAndLabels:
    model_outputs: Dict[str, Dict[str, torch.Tensor]]
    target_labels: Dict[str, Dict[str, torch.Tensor]]
    ids: Dict[str, Dict[str, List[str]]]


def filter_missing_outputs_and_labels(
    batch_ids: List[str],
    model_outputs: Dict[str, Dict[str, torch.Tensor]],
    target_labels: Dict[str, Dict[str, torch.Tensor]],
    missing_ids_info: MissingTargetsInfo,
    with_labels: bool = True,
) -> FilteredOutputsAndLabels:
    filtered_outputs = {}
    filtered_labels = {}
    filtered_ids: Dict[str, Dict[str, List[str]]] = {}

    precomputed = missing_ids_info.precomputed_missing_ids

    for output_name, output_inner_dict in model_outputs.items():
        filtered_inner_dict = {}
        filtered_inner_labels = {}
        filtered_inner_ids = {}

        output_missing_info = precomputed.get(output_name, {})
        if not output_missing_info:
            filtered_outputs[output_name] = output_inner_dict
            if with_labels:
                filtered_labels[output_name] = target_labels[output_name]
            else:
                filtered_labels[output_name] = {
                    inner_key: torch.tensor([]) for inner_key in output_inner_dict
                }
            filtered_ids[output_name] = {
                inner_key: batch_ids for inner_key in output_inner_dict
            }
            continue

        for inner_key, modality_output_tensor in output_inner_dict.items():
            cur_missing_ids = output_missing_info.get(inner_key, set())

            if not cur_missing_ids:
                filtered_inner_dict[inner_key] = modality_output_tensor
                if with_labels:
                    filtered_inner_labels[inner_key] = target_labels[output_name][
                        inner_key
                    ]
                else:
                    filtered_inner_labels[inner_key] = torch.tensor([])
                filtered_inner_ids[inner_key] = batch_ids
                continue

            valid_indices = [
                i for i, id_ in enumerate(batch_ids) if id_ not in cur_missing_ids
            ]
            valid_indices_tensor = torch.tensor(
                valid_indices,
                device=modality_output_tensor.device,
                dtype=torch.long,
            )

            output_tensor = modality_output_tensor.index_select(
                dim=0, index=valid_indices_tensor
            )
            filtered_inner_dict[inner_key] = output_tensor

            if with_labels:
                cur_targets = target_labels[output_name][inner_key].index_select(
                    dim=0, index=valid_indices_tensor
                )
                filtered_inner_labels[inner_key] = cur_targets
            else:
                filtered_inner_labels[inner_key] = torch.tensor([])

            ids = [batch_ids[i] for i in valid_indices]
            filtered_inner_ids[inner_key] = ids

        filtered_outputs[output_name] = filtered_inner_dict
        filtered_labels[output_name] = filtered_inner_labels
        filtered_ids[output_name] = filtered_inner_ids

    return FilteredOutputsAndLabels(
        model_outputs=filtered_outputs,
        target_labels=filtered_labels,
        ids=filtered_ids,
    )



import os
from copy import copy
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Tuple, Union

from torch import distributed as dist
from torch.nn.parallel import DistributedDataParallel

from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.models.model_setup_modules.meta_setup import al_meta_model
    from eir.setup.config import Configs

logger = get_logger(name=__name__)


def maybe_initialize_distributed_environment(
    configs: "Configs",
) -> Tuple["Configs", Union[int, None]]:
    is_distributed_run = in_distributed_env()
    if is_distributed_run:
        logger.info(
            "'LOCAL_RANK': '%s' environment variable found. Assuming "
            "distributed distributed training.",
            os.environ["LOCAL_RANK"],
        )
    else:
        return configs, None

    dist.init_process_group(backend="gloo")

    configs_copy = copy(configs)

    local_rank = int(os.environ["LOCAL_RANK"])
    if "cuda" in configs.gc.be.device:
        configs_copy.gc.be.device = f"cuda:{local_rank}"

    return configs_copy, local_rank


def maybe_make_model_distributed(
    device: str, model: "al_meta_model"
) -> "al_meta_model":
    if not in_distributed_env():
        return model

    local_rank = int(os.environ["LOCAL_RANK"])

    device_ids = None
    output_device = None

    if "cuda" in device:
        device_ids = [local_rank]
        output_device = local_rank

    model = AttrDelegatedDistributedDataParallel(
        module=model,
        device_ids=device_ids,
        output_device=output_device,
    )

    logger.info(
        "Initialized distributed model with rank '%d' and arguments "
        "'device_ids': '%s', 'output_device': '%s'.",
        local_rank,
        device_ids,
        output_device,
    )

    return model


def in_distributed_env() -> bool:
    if "LOCAL_RANK" in os.environ:
        return True
    return False


def in_master_node() -> bool:
    local_rank = int(os.environ.get("LOCAL_RANK", 0))
    global_rank = int(os.environ.get("RANK", 0))

    if local_rank == 0 and global_rank == 0:
        return True

    return False


def only_call_on_master_node(func: Callable):
    @wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        if in_master_node():
            result = func(*args, **kwargs)
            return result

        return

    return wrapper


class AttrDelegatedDistributedDataParallel(DistributedDataParallel):
    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)



from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Protocol

from eir.data_load.label_setup import al_label_transformers
from eir.setup import schemas
from eir.setup.output_setup_modules.array_output_setup import (
    ComputedArrayOutputInfo,
    set_up_array_output,
)
from eir.setup.output_setup_modules.image_output_setup import (
    ComputedImageOutputInfo,
    set_up_image_output,
)
from eir.setup.output_setup_modules.sequence_output_setup import (
    ComputedSequenceOutputInfo,
    set_up_sequence_output,
)
from eir.setup.output_setup_modules.tabular_output_setup import (
    ComputedTabularOutputInfo,
    set_up_tabular_output,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.input_setup import al_input_objects_as_dict


logger = get_logger(name=__name__)

al_output_objects = (
    ComputedTabularOutputInfo
    | ComputedSequenceOutputInfo
    | ComputedArrayOutputInfo
    | ComputedImageOutputInfo
)
al_output_objects_as_dict = Dict[str, al_output_objects]


def set_up_outputs_for_training(
    output_configs: schemas.al_output_configs,
    input_objects: Optional["al_input_objects_as_dict"] = None,
    target_transformers: Optional[Dict[str, al_label_transformers]] = None,
) -> al_output_objects_as_dict:
    all_outputs = set_up_outputs_general(
        output_configs=output_configs,
        setup_func_getter=get_output_setup_function_for_train,
        setup_func_kwargs={
            "input_objects": input_objects,
            "target_transformers": target_transformers,
        },
    )

    return all_outputs


class OutputSetupFunction(Protocol):
    def __call__(
        self,
        input_config: schemas.OutputConfig,
        **kwargs,
    ) -> al_output_objects: ...


class OutputSetupGetterFunction(Protocol):
    def __call__(self, output_config: schemas.OutputConfig) -> OutputSetupFunction: ...


def set_up_outputs_general(
    output_configs: schemas.al_output_configs,
    setup_func_getter: OutputSetupGetterFunction,
    setup_func_kwargs: Dict[str, Any],
) -> al_output_objects_as_dict:
    all_outputs = {}

    name_config_iter = get_output_name_config_iterator(output_configs=output_configs)

    for name, output_config in name_config_iter:
        setup_func = setup_func_getter(output_config=output_config)

        cur_output_data_config = output_config.output_info
        logger.info(
            "Setting up %s outputs '%s' from %s.",
            cur_output_data_config.output_name,
            cur_output_data_config.output_type,
            cur_output_data_config.output_source,
        )

        set_up_output = setup_func(output_config=output_config, **setup_func_kwargs)
        all_outputs[name] = set_up_output

    return all_outputs


def get_output_setup_function_for_train(
    output_config: schemas.OutputConfig,
) -> Callable[..., al_output_objects]:
    output_type = output_config.output_info.output_type

    mapping = get_output_setup_function_map()

    return mapping[output_type]


def get_output_setup_function_map() -> dict[str, Callable[..., al_output_objects]]:
    setup_mapping: dict[str, Callable[..., al_output_objects]] = {
        "tabular": set_up_tabular_output,
        "sequence": set_up_sequence_output,
        "array": set_up_array_output,
        "image": set_up_image_output,
    }

    return setup_mapping


def get_output_name_config_iterator(output_configs: schemas.al_output_configs):
    """
    We do not allow '.' as it is used in the weighted sampling setup.
    """

    for output_config in output_configs:
        cur_input_data_config = output_config.output_info
        cur_name = cur_input_data_config.output_name

        if "." in cur_name:
            raise ValueError(
                "Having '.' in the output name is currently not supported. Got '%s'."
                "Kindly rename '%s' to not include any '.' symbols.",
                cur_name,
                cur_name,
            )

        yield cur_name, output_config



import ast
import json
import operator
from collections import defaultdict
from functools import reduce
from pathlib import Path
from typing import Any, DefaultDict, Dict, Generator, Iterable, MutableMapping, Optional

import yaml

from eir.utils.logging import get_logger

logger = get_logger(name=__name__)


def get_yaml_iterator_with_injections(
    yaml_config_files: Iterable[str], extra_cl_args: Optional[list[str]]
) -> Generator[Dict, None, None]:
    if not extra_cl_args:
        yield from get_yaml_to_dict_iterator(yaml_config_files=yaml_config_files)
        return

    for yaml_config_file in yaml_config_files:
        loaded_yaml = load_yaml_config(config_path=yaml_config_file)

        yaml_file_path_object = Path(yaml_config_file)
        for extra_arg in extra_cl_args:
            extra_arg_parsed = extra_arg.lstrip("--")
            target_file, str_to_inject = extra_arg_parsed.split(".", 1)

            if target_file == yaml_file_path_object.stem:
                dict_to_inject = convert_cl_str_to_dict(str_=str_to_inject)

                logger.debug("Injecting %s into %s", dict_to_inject, loaded_yaml)
                loaded_yaml = recursive_dict_replace(
                    dict_=loaded_yaml, dict_to_inject=dict_to_inject
                )

        yield loaded_yaml


def convert_cl_str_to_dict(str_: str) -> dict:
    def _infinite_dict() -> DefaultDict:
        return defaultdict(_infinite_dict)

    infinite_dict = _infinite_dict()

    keys, final_value = str_.split("=", 1)
    keys_split = keys.split(".")

    try:
        final_value_parsed = ast.literal_eval(final_value)
    except (ValueError, SyntaxError):
        final_value_parsed = final_value

    inner_most_dict = reduce(operator.getitem, keys_split[:-1], infinite_dict)
    inner_most_dict[keys_split[-1]] = final_value_parsed

    dict_primitive = object_to_primitives(obj=infinite_dict)
    return dict_primitive


def get_yaml_to_dict_iterator(
    yaml_config_files: Iterable[str],
) -> Generator[dict, None, None]:
    for yaml_config in yaml_config_files:
        yield load_yaml_config(config_path=yaml_config)


def load_yaml_config(config_path: str) -> Dict[str, Any]:
    with open(config_path, "r") as yaml_file:
        config_as_dict = yaml.load(stream=yaml_file, Loader=yaml.FullLoader)

    return config_as_dict


def recursive_dict_replace(
    dict_: MutableMapping,
    dict_to_inject: MutableMapping,
) -> dict:
    for cur_key, cur_value in dict_to_inject.items():
        if cur_key not in dict_:
            dict_[cur_key] = {}

        old_dict_value = dict_.get(cur_key)
        cur_is_dict = isinstance(cur_value, MutableMapping)
        old_is_dict = isinstance(old_dict_value, MutableMapping)
        if cur_is_dict and old_is_dict:
            assert isinstance(cur_value, MutableMapping)
            assert isinstance(old_dict_value, MutableMapping)
            recursive_dict_replace(dict_=old_dict_value, dict_to_inject=cur_value)
        else:
            dict_[cur_key] = cur_value

    return dict(dict_)


def object_to_primitives(obj):
    return json.loads(json.dumps(obj, default=lambda o: o.__dict__))



from dataclasses import dataclass
from typing import Literal, Optional, Sequence, Union


@dataclass
class ImageOutputTypeConfig:
    """
    :param adaptive_normalization_max_samples:
        If using adaptive normalization (channel / element),
        how many samples to use to compute the normalization parameters.
        If None, will use all samples.

    :param resize_approach:
        The method used for resizing the images. Options are:
        - "resize": Directly resize the image to the target size.
        - "randomcrop": Resize the image to a larger size than the target and then
        apply a random crop to the target size.
        - "centercrop": Resize the image to a larger size than the target and then
        apply a center crop to the target size.

    :param mean_normalization_values:
        Average channel values to normalize images with. This can be a sequence matching
        the number of channels, or None. If None and using a pretrained model, the
        values used for the model pretraining will be used. If None and training from
        scratch, will iterate over training data and compute the running average
        per channel.

    :param stds_normalization_values:
        Standard deviation channel values to normalize images with. This can be a
        sequence mathing the number of channels, or None. If None and using a
        pretrained model, the values used for the model pretraining will be used.
        If None and training from scratch, will iterate over training data and compute
        the running average per channel.

    :param mode:
        An explicit mode to convert loaded images to. Useful when working with
        input data with a mixed number of channels, or you want to convert
        images to a specific mode.
        Options are
        - "RGB": Red, Green, Blue (channels=3)
        - "L": Grayscale (channels=1)
        - "RGBA": Red, Green, Blue, Alpha (channels=4)

    :param num_channels:
        Number of channels in the images. If None, will try to infer the number of
        channels from a random image in the training data. Useful when known
        ahead of time how many channels the images have, will raise an error if
        an image with a different number of channels is encountered.

    :param loss:
        Which loss to use for training the model. Either ``mse`` or ``diffusion``.

    :param diffusion_time_steps:
        Number of time steps to use for diffusion loss. Only used if ``loss`` is
        set to ``diffusion``.
    """

    size: Sequence[int] = (64,)
    resize_approach: Union[Literal["resize", "randomcrop", "centercrop"]] = "resize"
    adaptive_normalization_max_samples: Optional[int] = None
    mean_normalization_values: Union[None, Sequence[float]] = None
    stds_normalization_values: Union[None, Sequence[float]] = None
    mode: Optional[Literal["RGB", "L", "RGBA"]] = None
    num_channels: Optional[int] = None
    loss: Literal["mse", "diffusion"] = "mse"
    diffusion_time_steps: Optional[int] = 1000


@dataclass
class ImageOutputSamplingConfig:
    """
    :param manual_inputs:
        Manually specified inputs to use for sequence generation. This is useful
        if you want to generate sequences based on a specific input. Depending
        on the input type, different formats are expected:

        - ``sequence``: A string written directly in the ``.yaml`` file.
        - ``omics``: A file path to NumPy array of shape ``(4, n_SNPs)`` on disk.
        - ``image``: An image file path on disk.
        - ``tabular``: A mapping of :literal:`(column key: value)` written directly
          in the ``.yaml`` file.
        - ``array``: A file path to NumPy array on disk.
        - ``bytes``: A file path to a file on disk.


    :param n_eval_inputs:
        The number of inputs automatically sampled from the validation set for
        sequence generation.
    """

    manual_inputs: Sequence[dict[str, str]] = tuple()
    n_eval_inputs: int = 10



from functools import partial
from typing import Callable, Dict, Sequence, Union

from eir.experiment_io.experiment_io import load_serialized_input_object
from eir.predict_modules.predict_tabular_input_setup import (
    setup_tabular_input_for_testing,
)
from eir.setup import input_setup, schemas
from eir.setup.input_setup import al_input_objects_as_dict
from eir.setup.input_setup_modules import setup_array, setup_omics
from eir.setup.input_setup_modules.setup_bytes import ComputedBytesInputInfo
from eir.setup.input_setup_modules.setup_image import ComputedImageInputInfo
from eir.setup.input_setup_modules.setup_sequence import ComputedSequenceInputInfo
from eir.train_utils.step_logic import Hooks


def set_up_inputs_for_predict(
    test_inputs_configs: schemas.al_input_configs,
    ids: Sequence[str],
    hooks: Union["Hooks", None],
    output_folder: str,
) -> al_input_objects_as_dict:
    train_input_setup_kwargs = {
        "ids": ids,
        "output_folder": output_folder,
    }
    all_inputs = input_setup.set_up_inputs_general(
        inputs_configs=test_inputs_configs,
        hooks=hooks,
        setup_func_getter=get_input_setup_function_for_predict,
        setup_func_kwargs=train_input_setup_kwargs,
    )

    return all_inputs


def get_input_setup_function_for_predict(
    input_config: schemas.InputConfig,
) -> Callable[..., input_setup.al_input_objects]:
    mapping = get_input_setup_function_map_for_predict()
    input_type = input_config.input_info.input_type

    return mapping[input_type]


def get_input_setup_function_map_for_predict() -> (
    Dict[str, Callable[..., input_setup.al_input_objects]]
):
    setup_mapping: Dict[str, Callable[..., input_setup.al_input_objects]] = {
        "omics": setup_omics.set_up_omics_input,
        "tabular": setup_tabular_input_for_testing,
        "sequence": partial(
            load_serialized_input_object,
            input_class=ComputedSequenceInputInfo,
        ),
        "bytes": partial(
            load_serialized_input_object,
            input_class=ComputedBytesInputInfo,
        ),
        "image": partial(
            load_serialized_input_object,
            input_class=ComputedImageInputInfo,
        ),
        "array": partial(
            load_serialized_input_object,
            input_class=setup_array.ComputedArrayInputInfo,
        ),
    }

    return setup_mapping



from dataclasses import dataclass
from typing import Literal, Optional, Sequence

al_broker_projection_types = Literal[
    "lcl",
    "lcl_residual",
    "linear",
    "grouped_linear",
    "pool",
    "cnn",
    "interpolate",
]

al_broker_fusion_types = Literal["cross-attention", "sum", "cat+conv"]


@dataclass
class TensorMessageConfig:
    """
    :param name:
        Name of the message, used to identify the message in the broker, e.g. when
        a message is cached with a name, the name is used by other messages to
        use the cached tensor.

    :param layer_path:
        Path to the layer in the model that the message is extracted from.

    :param cache_tensor:
        Whether to cache the tensor in the broker.

    :param layer_cache_target:
        Whether to cache the input or output of the layer.

    :param use_from_cache:
        List of names of messages (i.e. from the 'name' field) that this message
        will use from the cache. Assumes that the messages have been cached earlier
        in the model flow.

    :param cache_fusion_type:
        Type of fusion to use when combining the cached tensors. Options are:

        - ``cross-attention``: Use cross-attention to combine the tensors.
        - ``sum``: Learnable gated sum to combine the tensors.
        - ``cat+conv``: Concatenate the tensors and apply a convolutional layer.

    :param projection_type:
        Type of projection to use when projecting the tensor to the target space.
        Options are:

        - ``lcl``: Locally connected layer.
        - ``lcl_residual``: Locally connected layer with residual connection.
        - ``cnn``: Convolutional layer, only supports down sampling for now.
        - ``linear``: Linear layer.
        - ``pool``: Adaptive average pooling layer.
        - ``grouped_linear``: Grouped linear layer (each dimension is projected
          separately with a learnable linear layer).

        If the tensor is already of the target size, no projection is performed.
    """

    name: str
    layer_path: str
    cache_tensor: bool = False
    layer_cache_target: Literal["input", "output"] = "output"
    use_from_cache: Optional[list[str]] = None
    cache_fusion_type: al_broker_fusion_types = "cat+conv"
    projection_type: al_broker_projection_types = "lcl"


@dataclass
class TensorBrokerConfig:
    """
    :param message_configs:
        List of message configurations for the broker.
    """

    message_configs: Sequence[TensorMessageConfig]



from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Sequence

import eir.models.output.sequence.sequence_output_modules
from eir.models.input.sequence.transformer_models import (
    BasicTransformerFeatureExtractorModelConfig,
    SequenceModelConfig,
)
from eir.models.output.sequence.sequence_output_modules import (
    SequenceOutputModuleConfig,
)
from eir.utils.logging import get_logger

if TYPE_CHECKING:
    from eir.setup.config import Configs

from eir.setup import config, schemas

logger = get_logger(name=__name__)


def get_configs_object_with_seq_output_configs(
    configs: "Configs",
) -> "Configs":
    seq_configs_iter = [asdict(i) for i in configs.output_configs]
    seq_configs = get_seq_output_configs(
        seq_configs=seq_configs_iter,
        base_output_configs=configs.output_configs,
    )

    input_configs = configs.input_configs
    extra_input_configs = converge_sequence_output_configs_to_input_configs(
        seq_output_configs=seq_configs,
        input_configs=input_configs,
    )
    # TODO: Filter the duplicated ones where we have original *and* extra
    seq_input_configs = list(input_configs) + list(extra_input_configs)

    seq_config_object_kwargs = configs.__dict__
    seq_config_object_kwargs["output_configs"] = seq_configs
    seq_config_object_kwargs["input_configs"] = seq_input_configs

    seq_config_object_kwargs.pop("gc")

    configs_object = config.Configs(**seq_config_object_kwargs)

    return configs_object


def get_seq_output_configs(
    seq_configs: Iterable[dict],
    base_output_configs: Iterable[schemas.OutputConfig],
) -> Sequence[schemas.OutputConfig]:
    config_objects = []

    for config_dict in seq_configs:
        cur_base_output_config = next(
            i
            for i in base_output_configs
            if i.output_info.output_name == config_dict["output_info"]["output_name"]
        )

        config_object = cur_base_output_config
        if cur_base_output_config.output_info.output_type == "sequence":
            config_object = init_seq_output_config(
                yaml_config_as_dict=config_dict,
                base_output_config=cur_base_output_config,
            )

        config_objects.append(config_object)

    return config_objects


def init_seq_output_config(
    yaml_config_as_dict: Dict[str, Any],
    base_output_config: schemas.OutputConfig,
) -> schemas.OutputConfig:
    """
    We include the input type here for more fine grained logic later when setting up
    the target and input configs.
    """

    seq_sampling_config_kwargs = yaml_config_as_dict.get("sampling_config", {})
    if seq_sampling_config_kwargs is None:
        seq_sampling_config_kwargs = {}

    seq_sampling_config = schemas.SequenceOutputSamplingConfig(
        **seq_sampling_config_kwargs
    )

    seq_config = schemas.OutputConfig(
        output_info=base_output_config.output_info,
        output_type_info=base_output_config.output_type_info,
        model_config=base_output_config.model_config,
        sampling_config=seq_sampling_config,
    )

    return seq_config


def converge_sequence_output_configs_to_input_configs(
    seq_output_configs: Sequence[schemas.OutputConfig],
    input_configs: Optional[Sequence[schemas.InputConfig]] = None,
) -> Sequence[schemas.InputConfig]:
    converged_input_configs = []

    for seq_output_config in seq_output_configs:
        if seq_output_config.output_info.output_type != "sequence":
            continue

        matched_input_config = None
        if input_configs:
            matched_input_configs = [
                i
                for i in input_configs
                if i.input_info.input_name == seq_output_config.output_info.output_name
            ]
            if matched_input_configs:
                assert len(matched_input_configs) == 1
                matched_input_config = matched_input_configs[0]
                logger.info(
                    "Found input config with matching name for sequence config: '%s'. "
                    "Will overload matching keys.",
                    seq_output_config.output_info.output_name,
                )

        cur_converged_input_config = _build_sequence_input_config_from_output(
            sequence_output_config=seq_output_config,
            input_config_to_overload=matched_input_config,
        )
        converged_input_configs.append(cur_converged_input_config)

    return converged_input_configs


def _build_sequence_input_config_from_output(
    sequence_output_config: schemas.OutputConfig,
    input_config_to_overload: Optional[schemas.InputConfig] = None,
) -> schemas.InputConfig:
    output_info = sequence_output_config.output_info
    input_info = schemas.InputDataConfig(
        input_source=output_info.output_source,
        input_type=output_info.output_type,
        input_name=output_info.output_name,
        input_inner_key=output_info.output_inner_key,
    )

    input_type_info = _build_matched_input_type_info(
        sequence_output_config=sequence_output_config
    )

    extracted_attributes = _build_matched_sequence_model_config_kwargs(
        sequence_output_config=sequence_output_config,
        input_config_to_overload=input_config_to_overload,
    )

    overloaded_model_config = schemas.SequenceModelConfig(
        model_type="eir-input-sequence-from-linked-output-default",
        model_init_config=extracted_attributes.model_init_config,
        **extracted_attributes.model_config_kwargs,
    )
    converged_input_config = schemas.InputConfig(
        input_info=input_info,
        input_type_info=input_type_info,
        model_config=overloaded_model_config,
        pretrained_config=extracted_attributes.pretrained_config,
    )

    return converged_input_config


def _build_matched_input_type_info(
    sequence_output_config: schemas.OutputConfig,
) -> schemas.SequenceInputDataConfig:
    output_type_info = sequence_output_config.output_type_info
    output_type_info_kwargs = output_type_info.__dict__

    input_type_info_keys = schemas.SequenceInputDataConfig.__dataclass_fields__.keys()
    matched_kwargs = {
        k: v for k, v in output_type_info_kwargs.items() if k in input_type_info_keys
    }

    input_type_info = schemas.SequenceInputDataConfig(**matched_kwargs)

    return input_type_info


@dataclass
class ExtractedAttributesFromOutputConfig:
    """
    To be used to build a matching input config.
    """

    model_config_kwargs: dict[str, Any]
    model_init_config: BasicTransformerFeatureExtractorModelConfig
    pretrained_config: Optional[schemas.BasicPretrainedConfig]


def _build_matched_sequence_model_config_kwargs(
    sequence_output_config: schemas.OutputConfig,
    input_config_to_overload: Optional[schemas.InputConfig] = None,
) -> ExtractedAttributesFromOutputConfig:
    sequence_keys_to_exclude = (
        _get_keys_to_exclude_from_output_input_sequence_overloading()
    )

    if input_config_to_overload:
        input_model_config = input_config_to_overload.model_config
        assert isinstance(
            input_model_config, (SequenceModelConfig, SequenceOutputModuleConfig)
        )

        model_config_kwargs = _extract_model_config_kwargs(
            model_config=input_model_config,
            keys_to_exclude=sequence_keys_to_exclude,
        )
        model_init_config = input_model_config.model_init_config
        pretrained_config = input_config_to_overload.pretrained_config
    else:
        output_model_config = sequence_output_config.model_config
        assert isinstance(
            output_model_config, (SequenceModelConfig, SequenceOutputModuleConfig)
        )

        model_config_kwargs = _extract_model_config_kwargs(
            model_config=output_model_config,
            keys_to_exclude=sequence_keys_to_exclude,
        )
        model_init_config_kwargs = output_model_config.model_init_config.__dict__
        model_init_config = BasicTransformerFeatureExtractorModelConfig(
            **model_init_config_kwargs
        )
        pretrained_config = None

    assert isinstance(model_init_config, BasicTransformerFeatureExtractorModelConfig)
    extracted_attributes = ExtractedAttributesFromOutputConfig(
        model_config_kwargs=model_config_kwargs,
        model_init_config=model_init_config,
        pretrained_config=pretrained_config,
    )

    return extracted_attributes


def _get_keys_to_exclude_from_output_input_sequence_overloading() -> tuple[str, ...]:
    """
    These keys are either exclusive to the output configuration, or possibly
    overloaded by the input configuration.
    """
    return (
        "model_type",
        "model_init_config",
        "projection_layer_type",
    )


def _extract_model_config_kwargs(
    model_config: (
        schemas.SequenceModelConfig
        | eir.models.output.sequence.sequence_output_modules.SequenceOutputModuleConfig
    ),
    keys_to_exclude: tuple[str, ...],
) -> dict[str, Any]:
    extracted = {
        k: v for k, v in model_config.__dict__.items() if k not in keys_to_exclude
    }
    return extracted



import math
from typing import Type

import torch
from torch import nn
from torch.nn import init as init

from eir.models.layers.norm_layers import LayerScale


def append_dims(tensor: torch.Tensor, target_shape_no_batch: list[int]) -> torch.Tensor:
    # exclude batch dimension
    current_dims = len(tensor.shape) - 1
    target_dims = len(target_shape_no_batch)
    if current_dims >= target_dims:
        return tensor

    new_shape = list(tensor.shape) + [1] * (target_dims - current_dims)
    return tensor.view(new_shape)


def retract_dims(
    tensor: torch.Tensor, target_shape_no_batch: list[int]
) -> torch.Tensor:
    """
    For example with:
        - tensor: [64, 3, 4]
        - target_shape_no_batch: [12]

        - current_dims = 3 - 1 = 2
        - target_dims = 1
        - Start flattening at dim 1

    We have 1 dimension per sample, so we flatten the last 2 dimensions to get:
        - [64, 12]
    """
    # exclude batch dimension
    current_dims = len(tensor.shape) - 1
    target_dims = len(target_shape_no_batch)

    if current_dims <= target_dims:
        return tensor

    start_dim = target_dims

    return tensor.flatten(start_dim=start_dim)


def _get_retracted_shape(input_shape: list[int], target_shape: list[int]) -> list[int]:
    if len(input_shape) <= len(target_shape):
        return input_shape

    n_dims_to_keep = len(target_shape) - 1
    dims_to_flatten = input_shape[n_dims_to_keep:]
    flattened_dim = math.prod(dims_to_flatten)

    retracted_shape = input_shape[:n_dims_to_keep] + [flattened_dim]

    return retracted_shape


def _apply_all_projections(
    x: torch.Tensor,
    projections: nn.ModuleDict,
    key_and_dim_order: list[tuple[str, int]],
) -> torch.Tensor:
    for key, dim in key_and_dim_order:
        x = _apply_projection(x=x, projection=projections[key], dim=dim)

    return x


def create_key_and_dim_order(projections: nn.ModuleDict) -> list[tuple[str, int]]:

    key_dim_pairs = []
    for key in projections.keys():
        dim_batch_padded = int(key.split("_")[1]) + 1
        key_dim_pairs.append((key, dim_batch_padded))

    sorted_pairs = sorted(key_dim_pairs, key=lambda x: x[1])

    return sorted_pairs


def _apply_projection(
    x: torch.Tensor,
    projection: nn.Module,
    dim: int,
) -> torch.Tensor:

    x = x.transpose(dim, -1)
    x = projection(x)
    x = x.transpose(dim, -1)
    return x


class GroupedLinearProjectionWrapper(nn.Module):
    def __init__(
        self,
        input_shape: torch.Size,
        target_shape: torch.Size,
        full_preactivation: bool = True,
        use_factorized_projection: bool = False,
    ):
        """
        Note: We sometimes get gradient explosions w/o the full_preactivation here,
        e.g. when sending tensors to the last layer in the network. The pre-activation
        seems to alleviate this somewhat.
        """
        super().__init__()
        self.input_shape = input_shape
        self.target_shape = target_shape
        self.full_preactivation = full_preactivation
        self.use_factorized_projection = use_factorized_projection

        input_shape_list = list(input_shape)
        target_shape_list = list(target_shape)

        self.projection: nn.Module
        identity_projection_class: Type[nn.Module]
        if len(input_shape) <= len(target_shape):
            self.projection = GroupedUpProjectionLayer(
                input_shape=input_shape_list,
                target_shape=target_shape_list,
            )
            identity_projection_class = GroupedUpProjectionLayer

        elif len(input_shape) > len(target_shape):
            down_class: Type[nn.Module]
            if self.use_factorized_projection:
                down_class = GroupedDownProjectionLayerFactorized
            else:
                down_class = GroupedDownProjectionLayer

            self.projection = down_class(
                input_shape=input_shape_list,
                target_shape=target_shape_list,
            )
            identity_projection_class = down_class
        else:
            raise ValueError()

        self.project_identity: nn.Module
        if input_shape == target_shape:
            self.project_identity = nn.Identity()
        else:
            self.project_identity = identity_projection_class(
                input_shape=input_shape_list,
                target_shape=target_shape_list,
            )

        self.norm = nn.LayerNorm(normalized_shape=input_shape, eps=1e-05)
        self.activation = nn.GELU()

        self.ls = LayerScale(dim=1, init_values=1e-5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.norm(x)

        if self.full_preactivation:
            identity = self.project_identity(out)
        else:
            identity = self.project_identity(x)

        out = self.projection(out)
        out = self.activation(out)

        out = self.ls(out)

        return out + identity


class GroupedUpProjectionLayer(nn.Module):
    def __init__(
        self,
        input_shape: list[int],
        target_shape: list[int],
    ):
        super().__init__()
        self.input_shape = input_shape
        self.target_shape = target_shape
        self.projections = _create_projections(
            input_shape=input_shape,
            target_shape=target_shape,
        )

        self.key_and_dim_order = create_key_and_dim_order(projections=self.projections)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = append_dims(tensor=x, target_shape_no_batch=self.target_shape)
        x = _apply_all_projections(
            x=x,
            projections=self.projections,
            key_and_dim_order=self.key_and_dim_order,
        )
        return x


class GroupedDownProjectionLayer(nn.Module):
    def __init__(
        self,
        input_shape: list[int],
        target_shape: list[int],
    ):
        super().__init__()
        self.input_shape = input_shape
        self.target_shape = target_shape

        retracted_shape = _get_retracted_shape(
            input_shape=self.input_shape,
            target_shape=target_shape,
        )
        self.projections = _create_projections(
            input_shape=retracted_shape,
            target_shape=target_shape,
        )

        self.key_and_dim_order = create_key_and_dim_order(projections=self.projections)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = retract_dims(tensor=x, target_shape_no_batch=self.target_shape)
        x = _apply_all_projections(
            x=x,
            projections=self.projections,
            key_and_dim_order=self.key_and_dim_order,
        )

        return x


class GroupedDownProjectionLayerFactorized(nn.Module):
    def __init__(
        self,
        input_shape: list[int],
        target_shape: list[int],
        full_preactivation: bool = True,
    ):
        super().__init__()
        self.input_shape = input_shape
        self.target_shape = target_shape
        self.full_preactivation = full_preactivation

        (
            self.factorized_shape,
            self.pre_match_projections,
        ) = get_pre_dim_matching_projections(
            input_shape=input_shape,
            target_shape=target_shape,
        )

        self.pre_match_key_and_dim_order = create_key_and_dim_order(
            projections=self.pre_match_projections,
        )

        retracted_shape = _get_retracted_shape(
            input_shape=self.factorized_shape,
            target_shape=target_shape,
        )

        self.downsample_identity = _create_projections(
            input_shape=retracted_shape,
            target_shape=target_shape,
        )
        self.downsample_identity_key_and_dim_order = create_key_and_dim_order(
            projections=self.downsample_identity,
        )

        self.projections = _create_projections(
            input_shape=retracted_shape,
            target_shape=target_shape,
        )
        self.projections_key_and_dim_order = create_key_and_dim_order(
            projections=self.projections,
        )

        self.norm_1 = nn.LayerNorm(normalized_shape=retracted_shape)
        self.act_1 = nn.GELU()

        self.ls = LayerScale(dim=1, init_values=1e-5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        bottleneck = _apply_all_projections(
            x=x,
            projections=self.pre_match_projections,
            key_and_dim_order=self.pre_match_key_and_dim_order,
        )
        bottleneck = retract_dims(
            tensor=bottleneck, target_shape_no_batch=self.target_shape
        )

        out = self.norm_1(bottleneck)

        if self.full_preactivation:
            identity = _apply_all_projections(
                x=out,
                projections=self.downsample_identity,
                key_and_dim_order=self.downsample_identity_key_and_dim_order,
            )
        else:
            identity = _apply_all_projections(
                x=bottleneck,
                projections=self.downsample_identity,
                key_and_dim_order=self.downsample_identity_key_and_dim_order,
            )

        out = _apply_all_projections(
            x=out,
            projections=self.projections,
            key_and_dim_order=self.projections_key_and_dim_order,
        )

        out = self.act_1(out)
        out = self.ls(out)

        return out + identity


def scaled_kaiming_uniform_(
    tensor,
    a=0,
    mode="fan_in",
    nonlinearity="leaky_relu",
    scale=1.0,
) -> torch.Tensor:
    fan = init._calculate_correct_fan(tensor, mode)
    gain = init.calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    bound = math.sqrt(3.0) * std * scale
    with torch.no_grad():
        return tensor.uniform_(-bound, bound)


class ScaledLinear(nn.Linear):
    def __init__(self, in_features, out_features, bias=True, scale=1.0):
        super().__init__(in_features, out_features, bias)
        self.scale = scale
        self.scaled_reset_parameters()

    def scaled_reset_parameters(self) -> None:
        scaled_kaiming_uniform_(self.weight, a=math.sqrt(5), scale=self.scale)
        if self.bias is not None:
            fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
            bound = (1 / math.sqrt(fan_in)) * self.scale if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)


def _create_projections(
    input_shape: list[int],
    target_shape: list[int],
) -> nn.ModuleDict:
    projections = nn.ModuleDict()

    n_missing_dims = len(target_shape) - len(input_shape)
    padded_input_shape = input_shape + [1] * n_missing_dims

    total_target_elements = math.prod(target_shape)

    input_and_target_dim_iter = zip(padded_input_shape, target_shape)
    for dim_index, (input_dim, target_dim) in enumerate(input_and_target_dim_iter):
        if input_dim == target_dim:
            continue

        layers: list[nn.Module] = []

        scale_factor = math.sqrt(target_dim / total_target_elements)
        linear_layer = ScaledLinear(
            in_features=input_dim,
            out_features=target_dim,
            scale=scale_factor,
        )

        layers.append(linear_layer)

        projections[f"dim_{dim_index}"] = nn.Sequential(*layers)

    return projections


def get_pre_dim_matching_projections(
    input_shape: list[int], target_shape: list[int]
) -> tuple[list[int], nn.ModuleDict]:
    factorized_shape = _calculate_factorized_shape(
        shape_to_factorize=input_shape,
        target_shape=target_shape,
    )

    projections = _create_projections(
        input_shape=input_shape,
        target_shape=factorized_shape,
    )

    return factorized_shape, projections


def _calculate_factorized_shape(
    shape_to_factorize: list[int], target_shape: list[int]
) -> list[int]:
    """
    shape_to_factorize: [64, 64, 16]
    target_shape: [2]
    n_dims_to_factorize = 3 - 1 = 2
    factorized = [64, 64, 16][:2 - 1] = [64]
    for dim in [64, 16]:
        factorized.append(int(sqrt(dim))

    Then we only factorize the last n dimensions not matching, applying int(sqrt())

    So we e.g. return [64, 8, 4]
    """

    n_dims_to_factorize = len(shape_to_factorize) - len(target_shape)
    cutoff_index = n_dims_to_factorize - 1

    factorized = shape_to_factorize[:cutoff_index]

    for dim in shape_to_factorize[cutoff_index:]:
        factorized.append(int(math.sqrt(dim)))

    return factorized
